import 'dart:convert';

import 'package:bird/smart_json.dart';
import 'package:dartz/dartz.dart' hide State;
import 'package:meta/meta.dart';

/// A class that can instantiate a prototype of itself and do a deep copy by cloning itself.
abstract class SmartClass<T extends SmartClass<T>> {

  /// Allows [SmartJSON] to instantiate itself without reflection.
  T creator();

  /// Create a new model from the current one with changes applied as specified by the instructions.
  T copyWith(T Function(T) instructions) => instructions(clone());

  /// Create a new model from the current one with changes applied as specified by the instructions.
  T copyApply(void Function(T) instructions) {
    final a = clone();
    instructions(a);
    return a;
  }

  /// Clones T.
  T clone();
}

abstract class PrototypeStringCodec<S> {
  S get defaultValue;

  Codec<S, String> stringCodec();
}

abstract class PrototypeCodec<S, T> {
  Codec<S, T> codec(S prototype);
}

abstract class SmartX<T extends SmartX<T, X>, X> extends Codec<T, X>
    with SmartClass<T> {

  const SmartX();

  /// Describes how this class should be serialized.
  PrototypeCodec<T, X> tactic();

  X to() => encode(this as T);

  /// Errors need to be handled in the tactic.
  T from(X encodedData) => decode(encodedData);

  @override
  Converter<T, X> get encoder =>
      AnonymousConverter(tactic()
          .codec(creator())
          .encode);

  @override
  Converter<X, T> get decoder =>
      AnonymousConverter(tactic()
          .codec(creator())
          .decode);

  /// Clones T.
  @override
  T clone() => creator().from(to());
}

/// JSON parsing based on the Prototype Design Pattern.
///
/// A way to parse json with the goal of avoiding
/// type errors and key redundancy.
/// The main goal is to reduce boilerplate and accidental errors.
///
/// 1. Provide variables that extend [JSONSmartVariableInterface].
/// 2. Implement [creator] where a new instance of the class is created.
/// 3. Implement a [JSONPrototypeCodec].
///
/// Invalid Strings passed to [fromStringJson] will be discarded and the initial
/// implementation (with it's default values) will be returned.
///
/// Although marked as immutable, the SmartVariables inside a class are mutable.
/// Always call [clone] before mutating those variables,
@immutable
abstract class SmartJSON<T extends SmartJSON<T>> extends SmartX<T, dynamic>
    implements PrototypeStringCodec<T> {

  const SmartJSON();

  /// Describes how this class should be serialized.
  ///
  /// Casts RawSerialization tactic to one that provides a toString codec.
  @override
  JSONPrototypeCodec<T> tactic();

  CrystalizedCodec<T, String> convert() =>
      CrystalizedCodec<T, String>.fromCodec(
        instance: this as T,
        codec: stringCodec(),
      );

  @override
  Codec<T, String> stringCodec() =>
      this.fuse(tactic().toStringCodec(defaultValue));

  @override
  T get defaultValue => this as T;

  /// Return the JSON representation of this Object encoded as a String.
  @override
  String toString() => convert().to();

  String debugToString() =>
      "${runtimeType.toString()}(toString(): ${toString()})";

  @override
  bool operator ==(Object other) =>
      identical(this, other) || runtimeType == other.runtimeType
          && toString() == other.toString();

  @override
  int get hashCode => toString().hashCode;
}

abstract class SmartVariableInterface<T, ENCODED>
    implements PrototypeStringCodec<T> {

  Option<T> _val;

  Option<T> get option => _val;

  void set(Option<T> t) => _val = t;

  /// A setter for [val] without having to import dartz.
  void safeSet(T t) => set(optionOf(t));

  /// An apply setter for [val] to make the use not have to import dartz.
  void safeApplySet(T Function(T) t) => safeSet(t(nullVal()));

  /// The current value is passed and it can be mapped to a different value.
  U apply<U>(U Function(T) a) => a(nullVal());

  /// Safe version of [apply].
  U safeApply<U>(U Function() ifNull, U Function(T) a) => _val.fold(ifNull, a);

  /// A getter for [_val] without having to import dartz.
  ///
  /// Returns null when this value is not set.
  T nullVal() => _val.getOrElse(() => null);

  ENCODED to();
}

/// Variables that can serialize themselves and be
/// deserialized.
///
/// A bunch of data types are already implemented.
///
/// Check out their documentation for more info.
abstract class JSONSmartVariableInterface<T>
    extends SmartVariableInterface<T, Object> {

  /// The current value. Should never be null.
  /// Use [optionOf] to set it.
  @override
  Option<T> _val;

  /// This describes how this value can be serialized.
  Codec<Object, T> codec();

  @override
  Codec<T, String> stringCodec() => codec().inverted.fuse(json);

  @override
  T get defaultValue => nullVal();

  @override
  Object to() => codec().decode(nullVal());

  JSONSmartVariableInterface(T init) {
    safeSet(init);
  }

  @override
  String toString() => json.encode(to());
}

/// Returns the value in [list] that equals its toString method to
/// the string of the returned function. If [list] doesn't contain
/// a value which corresponds to [c] then the initial value [init]
/// is returned.
///
/// Useful for converting strings to Enums.
U Function(dynamic c) firstOfList<U>(List<U> list, U init) {
  return (dynamic c) {
    return list.firstWhere((a) {
      return a.toString() == c.toString();
    }, orElse: () => init);
  };
}

class AnonymousConverter<S, T> extends Converter<S, T> {

  final T Function(S) _convert;

  const AnonymousConverter(this._convert);

  @override
  T convert(S input) => _convert(input);
}

class AnonymousCodec<S, T> extends Codec<S, T> {

  final Converter<T, S> _decoder;
  final Converter<S, T> _encoder;

  const AnonymousCodec(this._encoder, this._decoder);

  @override
  Converter<T, S> get decoder => _decoder;

  @override
  Converter<S, T> get encoder => _encoder;
}

class CrystalizedCodec<F, T> {

  final T Function() to;
  final F Function(T t) from;

  const CrystalizedCodec({
    @required this.to,
    @required this.from,
  });

  CrystalizedCodec.fromCodec({
    @required F instance,
    @required Codec<F, T> codec,
  })
      :
        this.to = (() => codec.encode(instance)),
        this.from = ((T t) => codec.decode(t))
  ;
}

class FallbackCodecConfiguration<S, T> {

  final T fallbackEncodedValue;
  final S fallbackDecodedValue;
  final String debugLabel;
  final String loggerName;

  const FallbackCodecConfiguration(this.fallbackEncodedValue,
      this.fallbackDecodedValue,
      {this.debugLabel = "", this.loggerName = ""});

  const FallbackCodecConfiguration.nulll(
      {this.debugLabel = "", this.loggerName = ""})
      : fallbackDecodedValue = null,
        fallbackEncodedValue = null;
}

class FallbackCodec<S, T> extends Codec<S, T> {

  final Codec<S, T> codec;
  final FallbackCodecConfiguration<S, T> configuration;

  const FallbackCodec(this.codec, this.configuration);

  @override
  Converter<S, T> get encoder {
    return AnonymousConverter((S s) {
      try {
        return codec.encode(s);
      } catch (e) {
        print("FallbackCodec: Failed encoding '$S' to '$T' with $codec. '${configuration
                        .debugLabel}'");
//        Logger(configuration.loggerName).log(
//            Level.WARNING,
//                );
        return configuration.fallbackEncodedValue;
      }
    });
  }

  @override
  Converter<T, S> get decoder {
    return AnonymousConverter((T t) {
      try {
        return codec.decode(t);
      } catch (e) {
        print("FallbackCodec: Failed decoding '$T' to '$S' with $codec. '${configuration
                        .debugLabel}'");
//        Logger(configuration.loggerName).log(
//            Level.WARNING,
//            );
        return configuration.fallbackDecodedValue;
      }
    });
  }
}