import 'dart:convert';

import 'package:bird/smart_json.dart';

// All these seem like they could be created from primitives.
// Maybe try that someday. This should reduce error sources and ease
// the expandability of types to more complex ones by providing
// an example.

/// A Map of T
///
/// {
///   key : {
///     "a" : T
///   }
/// }
class SmartDictT<T> extends JSONSmartVariableInterface<Map<String, T>> {

  final T Function(dynamic) tCreator;

  SmartDictT(Map<String, T> init, {
    this.tCreator,
  })
      : super(init);

  @override
  Codec<dynamic, Map<String, T>> codec() {
    return IsJSONSmartCodec(
      fromJSON: (dynamic a) =>
          Map<String, dynamic>
              .from(a as Map)
              .map(
                (String a, dynamic b) {
              if (tCreator != null) {
                return MapEntry(a, tCreator(b));
              } else {
                return MapEntry(a, b as T);
              }
            },
          ),
    );
  }
}

/// A map of SmartJSON models
///
/// {
///   "key" : SmartJSON<T>
/// }
class SmartModelDict<T extends SmartJSON<T>>
    extends JSONSmartVariableInterface<Map<String, T>> {

  final T Function(dynamic) tCreator;

  SmartModelDict(Map<String, T> init, T Function() creator)
      : assert(creator != null),
        tCreator = ((dynamic a) {
          final c = creator();
          assert(c != null);
          return c.from(a);
        }),
        super(init);

  @override
  Codec<dynamic, Map<String, T>> codec() {
    return IsJSONSmartCodec(
      fromJSON: (dynamic a) {
        return Map<String, dynamic>
            .from(a as Map<String, dynamic>)
            .map(
              (String a, dynamic b) => MapEntry(a, tCreator(b)),
        );
      },
    );
  }
}

/// A Map of Strings mapped to a List of T.
///
/// {
///   key : {
///     "a": [T, T, T],
///   }
/// }
class SmartRegistry<T> extends SmartDictT<List<T>> {

  SmartRegistry(Map<String, List<T>> init,
      {
        T Function(dynamic) tCreator,
      })
      : super(
    init,
    tCreator: tCreator == null
              ? (dynamic a) => List.from(a as List)
              :
              (dynamic a) =>
                  List<dynamic>.from(a as List).map(tCreator).toList(),
  );

  void applyToggle(String key, T t) {
    safeApplySet((r) {
      r.putIfAbsent(key, () => []);
      if (r[key].contains(t)) {
        r[key].remove(t);
      } else {
        r[key].add(t);
      }
      return r;
    });
  }

  void applyAdd(String key, T t) {
    safeApplySet((r) {
      r.putIfAbsent(key, () => []);
      if (!r[key].contains(t)) {
        r[key].add(t);
      }
      return r;
    });
  }

  void applyAddAll(String key, List<T> t) {
    safeApplySet((r) {
      r.putIfAbsent(key, () => []);
      t.forEach((val) {
        if (!r[key].contains(val)) {
          r[key].add(val);
        }
      });
      return r;
    });
  }

  void applyAddOrRemove(String key, T t, bool add) {
    safeApplySet((r) {
      r.putIfAbsent(key, () => []);
      if (add) {
        if (!r[key].contains(t)) {
          r[key].add(t);
        }
      } else {
        if (r[key].contains(t)) {
          r[key].remove(t);
        }
      }
      return r;
    });
  }
}

/// A Map that contains maps for each key with T as values of the inner map
///
/// {
///   key: {
///     "a" : {
///       "aa" : T
///     }
///   }
/// }
class SmartDomainIdValue<T> extends SmartDictT<Map<String, T>> {
  SmartDomainIdValue(Map<String, Map<String, T>> init,
      {T Function(dynamic) tCreator})
      : super(init,
    tCreator: (dynamic a) =>
        Map<String, dynamic>.from(a as Map).map((String a, dynamic b) {
          if (tCreator != null) {
            return MapEntry(a, tCreator(b));
          } else {
            return MapEntry(a, b as T);
          }
        }),
  );
}