import 'package:bird/smart_json.dart';

/// Can be used to implement unusual inheritance hierarchies.
/// Take a look at [EthereumKeystoreCryptoModel] for an example.
///
/// If no fixed hierarchy is set then you might want to use [SmartInherit].
class SmartRedirect<U> extends SmartJSON<SmartRedirect<U>> {

  final SmartDynamic data = SmartDynamic(null);

  final U Function(dynamic d) produce;

  SmartRedirect(this.produce);

  static SmartRedirect<U> type<U extends InheritableSmartJSON<U>>(U deflt,
      List<U> subclasses) {
    return SmartRedirect<U>((dynamic map) {
      if (map is Map<String, dynamic>) {
        final String type = map["type"] as String;

        if (type == null) {
          return deflt;
        }

        final mappedValue = subclasses.firstWhere((a) =>
        a.typeSerializationString() == type, orElse: () => null);

        if (mappedValue == null) {
          return deflt;
        }

        return mappedValue.from(map);
      }
      return deflt;
    });
  }

  U get access => produce(data.nullVal());

  @override
  SmartRedirect<U> creator() => SmartRedirect<U>(produce);

  @override
  JSONPrototypeCodec<SmartRedirect<U>> tactic() => JSONSingleTactic(data);
}

/// Use this to serialize an inheritance hierarchy.
///
/// It stored the type of the subclass as an entry inside it's map under the "type" key.
/// Therefore "type" should not be used as a key.
///
/// Provide an abstract class that inherits from [InheritableSmartJSON] for each subclass to inherit.
// Note: Performance improvements possible by replacing [subclasses] with a map (N to 1)
class SmartInherit<T extends InheritableSmartJSON<T>>
    extends SmartTFrom<SmartRedirect<T>> {
  SmartInherit(T deflt, List<T> subclasses)
      : super(SmartRedirect.type(deflt, subclasses));

  void setData(T t) {
    nullVal().data.safeSet(t.to());
  }
}

abstract class InheritableSmartJSON<T extends SmartJSON<T>>
    extends SmartJSON<T> {
  @override
  JSONPrototypeCodec<T> tactic() {
    final more = mapTacticVars();

    assert(() {
      if (more.containsKey("type")) {
        throw Exception(
            "The key 'type' cannot be used as a variable inside of subclasses of $InheritableSmartJSON.");
      }
      return true;
    }());

    return JSONMapTactic.ignoreMissing(
        {"type": SmartString(typeSerializationString())}..addAll(more));
  }

  String typeSerializationString() => runtimeType.toString();

  Map<String, JSONSmartVariableInterface> mapTacticVars();
}
