import 'package:bird/base.dart';
import 'package:bird/bird.dart';
import 'package:bird/smart_json.dart';
import 'package:meta/meta.dart';

class WaveStorageTyped<T extends PrototypeStringCodec<U>, U> {
    final T deflt;

    const WaveStorageTyped(this.deflt);

    SmartWaveStorage<T, U> at(StorageBaseAt<String> storage) {
        return WaveStorage.prototype<T, U>(storageAt: storage, deflt: deflt);
    }
}

abstract class WaveStorage<T> implements InitBloc {
    static SmartWaveStorage<T, U> prototype<T extends PrototypeStringCodec<U>, U>({
        @required StorageBaseAt<String> storageAt,
        @required T deflt,
        FallbackCodecConfiguration<U, String> Function(U) fallbackConfiguration,
        void Function(WaveStorage<U>) onInit,
    }) {
        final defltValue = deflt.defaultValue;
        final fallbackConfig = fallbackConfiguration?.call(defltValue) ??
            FallbackCodecConfiguration("", defltValue,
                debugLabel: "Location: ${storageAt.location}");

        return SmartWaveStorage<T, U>(
            deflt,
            storageAt,
            FallbackCodec(deflt.stringCodec(), fallbackConfig).decode,
            FallbackCodec(deflt.stringCodec(), fallbackConfig).encode,
            defltValue,
            onInit: onInit,
        );
    }

    /// Provides a wave of the data.
    Wave<T> get dataWave;

    /// Call this to set new data.
    Future<void> set(T t);

    /// Tells whether this value has been set since the app is running.
    BoolWave wasMutated;
}

class SmartWaveStorage<U extends PrototypeStringCodec<T>, T> extends WaveStorageImpl<T> {
    final U smartInit;

    SmartWaveStorage(this.smartInit,
        StorageBaseAt<String> storage,
        T Function(String str) convertFromString,
        String Function(T t) convertToString,
        T initial,
        {void Function(WaveStorage<T>) onInit})
        : super(storage, convertFromString, convertToString, initial,
        onInit: onInit);

    SignalLens<T> get lens => SignalLens(set, dataWave);
}

/// Implementation of a [WaveStorage].
///
/// Important: If the storage returns null values then this is interpreted as
/// 'the value does not exist'.
class WaveStorageImpl<T> implements WaveStorage<T> {
    final StorageBaseAt<String> storage;
    final T Function(String str) convertFromString;
    final String Function(T t) convertToString;
    final T _initial;

    bool _isInitialized;
    Signal<bool> _wasSetSomeTime;
    Signal<T> _data;

    WaveStorageImpl(this.storage, this.convertFromString, this.convertToString, this._initial,
        {void Function(WaveStorage<T>) onInit}) {
        _data = Signal(_initial, onListen: _onListen);
        _isInitialized = false;
        _wasSetSomeTime = Signal(false);
        onInit?.call(this);
        wasMutated = _wasSetSomeTime.wave.as(BoolWave.from);
    }

    Future<void> _onListen() async {
        if (!_isInitialized) {
            final storedData = await storage.get();
            if (!_data.isClosed && storedData != null && storedData.isNotEmpty) {
                _data.add(convertFromString(storedData));
                _wasSetSomeTime.add(true);
            } else {
                _data.add(_initial);
            }
            _isInitialized = true;
        }
    }

    @override
    BoolWave wasMutated;

    @override
    Future<void> set(T t) async {
        _wasSetSomeTime.add(true);
        await storage.set(convertToString(t));
        _data.add(t);
    }

    @override
    Wave<T> get dataWave => _data.wave;

    @override
    Future<void> init() async => _onListen();

    @override
    Future<void> dispose() async {
        _data.close();
        _wasSetSomeTime.close();
    }
}
