import 'package:bird/smart_api.dart';
import 'package:dartz/dartz.dart';
import 'package:meta/meta.dart';
import 'package:dartz/dartz.dart' hide State;

const UnknownAPIError unknownAPIError = UnknownAPIError(None(), None());

UnknownAPIError unknownHintedAPIError(String description,
        {StackTrace stacktrace}) =>
    UnknownAPIError(optionOf(description), optionOf(stacktrace));

/// It is not known why the error occurred.
///
/// Neither [description] nor [stacktrace] go into the identity of this object.
class UnknownAPIError extends APIErrors {
  /// Most of the time Exception.toString().
  final Option<String> description;

  /// The stacktrace for the error.
  final Option<StackTrace> stacktrace;

  const UnknownAPIError(this.description, this.stacktrace);

  @override
  String toString() =>
      'UnknownAPIError{description: $description, stacktrace: $stacktrace';

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
      other is UnknownAPIError && runtimeType == other.runtimeType;

  @override
  int get hashCode => 0;
}

const APIResponseCustomErrorDetectedAPIError vendorError =
    APIResponseCustomErrorDetectedAPIError();

/// ------- API Response
///
/// API responses can have error codes and error messages
/// if an api provides custom error messages and one of such
/// messages were found then use this error.
class APIResponseCustomErrorDetectedAPIError extends APIErrors {
  const APIResponseCustomErrorDetectedAPIError();

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
      other is APIResponseCustomErrorDetectedAPIError &&
          runtimeType == other.runtimeType;

  @override
  int get hashCode => 0;
}

APIUnexpectedResponseError unexpectedResponseError(
        ResponseInformation response) =>
    APIUnexpectedResponseError(response);

/// When the response did not meet a criteria then this error should be thrown.
///
/// E.g. when the content type did not match up but should have this can be returned.
///
/// Be careful for == and hashcode only the type counts.
class APIUnexpectedResponseError extends APIErrors {
  final ResponseInformation response;

  const APIUnexpectedResponseError(this.response);

  @override
  String toString() => 'APIUnexpectedResponseError{response: $response}';

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
      other is APIUnexpectedResponseError && runtimeType == other.runtimeType;

  @override
  int get hashCode => 0;
}

class ResponseInformation {
  final Option<int> status;
  final Option<String> data;
  final Option<List<String>> contentType;

  const ResponseInformation({
    @required this.status,
    @required this.data,
    @required this.contentType,
  });

  const ResponseInformation.none()
      : status = const None(),
        data = const None(),
        contentType = const None();

  @override
  String toString() => 'ReponseInformation{'
      'status: ${status.map((a) => a.toString()).getOrElse(() => "/")}, '
      'data: ${data.getOrElse(() => "/")}, '
      'contentType: ${contentType.getOrElse(() => ["/"])}}';

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
      other is ResponseInformation &&
          runtimeType == other.runtimeType &&
          status == other.status &&
          data == other.data &&
          contentType == other.contentType;

  @override
  int get hashCode => status.hashCode ^ data.hashCode ^ contentType.hashCode;
}

const TimeoutAPIError timeoutError = TimeoutAPIError();

/// If a running request has timed out pass this error.
class TimeoutAPIError extends APIErrors {
  const TimeoutAPIError();

  @override
  bool operator ==(Object other) =>
      identical(this, other) || other is TimeoutAPIError;

  @override
  int get hashCode => 0;
}

APIStatusError statusError(int status, [String body]) =>
    APIStatusError(optionOf(status), optionOf(body));

/// This is for status errors e.g. 400, 413, ...
class APIStatusError extends APIErrors {
  final Option<int> statusCode;

  final Option<String> body;

  APIStatusError(this.statusCode, this.body);

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
      other is APIStatusError &&
          runtimeType == other.runtimeType &&
          statusCode == other.statusCode;

  @override
  int get hashCode => statusCode.hashCode;

  @override
  String toString() => '$APIStatusError{statusCode: $statusCode, body: $body}';
}

const NoInternetAPIError noInternetError = NoInternetAPIError();

/// ------- API Request
///
/// "No host found" message that indicates
/// no internet connection available.
class NoInternetAPIError extends APIErrors {
  const NoInternetAPIError();

  @override
  bool operator ==(Object other) =>
      identical(this, other) || other is NoInternetAPIError;

  @override
  int get hashCode => 0;
}

const LocalNotEnoughCreditsAvailableAPIError noCreditsError =
    LocalNotEnoughCreditsAvailableAPIError();

/// ------- Credits
///
/// Pass this when there are not enough credits found in the credit manager to execute the call
class LocalNotEnoughCreditsAvailableAPIError extends APIErrors {
  const LocalNotEnoughCreditsAvailableAPIError();

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
      other is LocalNotEnoughCreditsAvailableAPIError &&
          runtimeType == other.runtimeType;

  @override
  int get hashCode => 0;
}

const CancelledAPIError cancelledError = CancelledAPIError();

/// If a request was either cancelled by the user or the server
class CancelledAPIError extends APIErrors {
  const CancelledAPIError();

  @override
  bool operator ==(Object other) =>
      identical(this, other) || other is CancelledAPIError;

  @override
  int get hashCode => 0;
}
