import 'package:bird/smart_api.dart';
import 'package:bird/smart_json.dart';
import 'package:dartz/dartz.dart';

/// Error objects for describing errors.
///
/// All errors inherit [AllErrors]. These errors describe problems that
/// arise from handling local data and working with APIs.
///
/// Consider creating more errors if a common error was not implemented
/// and if it could help the user resolve a problem they are having.
/// If you add more errors then don't forget to also add that error to the
/// mapping function of the super class.
///
/// # [StorageErrors]
/// These errors describe errors that can happen when data is accessed locally.
///
/// - [FailedParsingStorageError] for when e.g. JSON data is loaded from disk
///    and needs to be parsed but it couldn't (if the JSON was invalid for example).
/// - [DataNotFoundStorageError] for when data was expected to be present somewhere but does
///    not exist at the expected location.
/// - [UnknownStorageError] If the above errors do not describe the error then use this one.
///
///
/// # [APIErrors]
/// These errors describe errors that arise when dealing with APIs from an external location.
///
/// - [CancelledAPIError] The request was cancelled by the user.
/// - [LocalNotEnoughCreditsAvailableAPIError] APIs may have a limit. Different policies
///    can limit the usage of certain methods. If it is calculated locally that an API
///    path has run out of credits then this should be used.
/// - [NoInternetAPIError] There is no certain way to determine if an API call was cancelled
///    because there was no internet but a somewhat sure way to do that is to check for a
///    "No host found" message inside the error signature. Use this error when doing that.
/// - [APIStatusError] Non 2XX status messages should be put here.
/// - [TimeoutAPIError] For when request times out.
/// - [APIResponseCustomErrorDetectedAPIError] Some response data contain custom error
///    messages that are specified by the vendor. These errors should be put here.
/// - [UnknownAPIError] If none of the above errors describe the error then use this one.
///
abstract class AllErrors with SubUnion2<APIErrors, StorageErrors> {
  const AllErrors._();
}

/// Errors than can occur while loading data from an API.
abstract class APIErrors extends AllErrors
    with
        SubSubUnion8<
            LocalNotEnoughCreditsAvailableAPIError,
            NoInternetAPIError,
            UnknownAPIError,
            TimeoutAPIError,
            CancelledAPIError,
            APIResponseCustomErrorDetectedAPIError,
            APIStatusError,
            APIUnexpectedResponseError> {
  const APIErrors() : super._();

  // ignore: use_to_and_as_if_applicable
  Either<APIErrors, T> left<T>() => Left(this);
}

/// Errors than can occur while loading data from local storage.
abstract class StorageErrors extends AllErrors
    with
        SubSubUnion3<DataNotFoundStorageError, FailedParsingStorageError,
            UnknownStorageError> {
  const StorageErrors() : super._();

  Either<StorageErrors, T> asLeft<T>() => Left(this);
}
