import 'package:bird/smart_api.dart';
import 'package:dartz/dartz.dart';
import 'package:meta/meta.dart';

/// A single unit of state management inside socket channels.
///
/// Reducers can define which message (at a socket state) leads to which channel state.
///
/// That way a channel can easily define behaviors for a wide variety of implementations.
///
/// Example:
///     - A channel may connect/disconnect on specific messages
///     - parse and pass a message to a callback.
///
/// Each reducer can hold state and will be disposed when the channel  is disposed.
abstract class SocketChannelReducer {
  ReducerStateInterface reducerStateInterface;

  void register(ReducerStateInterface i) => this.reducerStateInterface = i;

  /// Provide a state transition function for various messages at a specific socket state.
  ///
  /// Will be called when a new message is received and only a snapshot of the current state is passed.
  /// That means not every state change will be passed but every message.
  SocketReducerAction messageReducer(
      APIState currentState, StringConversion message);

  /// Will be disposed when the socket connection itself is disposed.
  void dispose();
}

abstract class SocketReducerAction {
  const SocketReducerAction();

  T map<T>({
    @required T Function() identity,
    @required T Function() connected,
  }) {
    if (this is SocketReducerActionIdentity) {
      return identity();
    } else if (this is SocketReducerActionConnected) {
      return connected();
    } else {
      throw Exception("Not implemented");
    }
  }
}

class SocketReducerActionIdentity extends SocketReducerAction {
  const SocketReducerActionIdentity();
}

class SocketReducerActionConnected extends SocketReducerAction {
  const SocketReducerActionConnected();
}

/// Implemented by classes that handle reducers.
abstract class ReducerStateInterface {
  /// Sends a message to the raw socket.
  void sendMessage(String msg);

  /// Disconnects the implementer.
  void disconnect(AllErrors error);
}

/// Prints on every reducer call. Acts as an identity reducer and doesn't change anything.
class DebugSocketChannelReducer extends SocketChannelReducer {
  final String channelName;

  DebugSocketChannelReducer(this.channelName);

  @override
  void register(ReducerStateInterface i) {
    print("$DebugSocketChannelReducer register @ $channelName register called");
  }

  @override
  SocketReducerAction messageReducer(
      APIState currentState, StringConversion message) {
    print(
        "$DebugSocketChannelReducer messageReducer @ $channelName at state $currentState with message ${message.str}");
    return const SocketReducerActionIdentity();
  }

  @override
  void dispose() => print("Disposing reducer @ $channelName");
}

/// Used to change a state to connected when a certain message was received.
///
/// Will only act on [APIStateLoading] states and approved messages by [becomeConnectedMessage].
class ConnectedChannelReducer extends SocketChannelReducer {
  final bool Function(StringConversion str) becomeConnectedMessage;

  ConnectedChannelReducer(this.becomeConnectedMessage);

  @override
  SocketReducerAction messageReducer(
      APIState currentState, StringConversion message) {
    if (currentState == const APIStateLoading() &&
        becomeConnectedMessage(message)) {
      return const SocketReducerActionConnected();
    } else {
      return const SocketReducerActionIdentity();
    }
  }

  @override
  void dispose() {}
}

///// Reacts to any message and sets the state to disconnected when no messages
///// were received in [period] time.
/////
///// Can also act as a passive heartbeat listener.
/////
///// If you need to react to received heartbeats then use [MarcopoloChannelReducer]
/////
///// It will only act on [APIStateConnected] states.
//class TimeoutChannelReducer extends SocketChannelReducer {
//
//    @visibleForTesting
//    Timer t;
//
//    final Duration period;
//
//    final AllErrors disconnectError;
//
//    TimeoutChannelReducer({
//        @required this.period,
//        this.disconnectError,
//    });
//
//    @override
//    SocketReducerAction messageReducer(APIState currentState, StringConversion message) {
//        t?.cancel();
//
//        if (currentState is APIStateConnected) {
//            t = Timer(
//                period,
//                    () {
//                    if (reducerStateInterface == null)
//                        throw Exception("You forgot to call 'register' on the reducer");
//
//                    reducerStateInterface.disconnect(
//                        disconnectError ?? unknownHintedAPIError("Did not receive any messages in ${period.inMilliseconds} ms"));
//                },
//            );
//        }
//
//        return const SocketReducerActionIdentity();
//    }
//
//    @override
//    void dispose() async => t.cancel();
//}

/// Reacts to messages by sending back a message to the socket.
///
/// Can be used for incoming heartbeats or ping/pong schemes.
///
/// It will only act on [APIStateConnected] states.
class MarcopoloChannelReducer<T> extends SocketChannelReducer {
  final Option<String> Function(StringConversion) _messageToSend;

  MarcopoloChannelReducer(this._messageToSend);

  @override
  SocketReducerAction messageReducer(
      APIState currentState, StringConversion message) {
    if (currentState is APIStateConnected) {
      if (reducerStateInterface == null) {
        throw Exception("You forgot to call 'register' on the reducer");
      }

      _messageToSend(message).fold(() {}, reducerStateInterface.sendMessage);
    }

    return const SocketReducerActionIdentity();
  }

  @override
  void dispose() {}
}

class ModelCallbackChannelReducer<T> extends SocketChannelReducer {
  final void Function(T t) callback;

  final T Function(StringConversion) converter;

  ModelCallbackChannelReducer({
    @required this.callback,
    @required this.converter,
  });

  @override
  SocketReducerAction messageReducer(
      APIState currentState, StringConversion message) {
    callback(converter(message));
    return const SocketReducerActionIdentity();
  }

  @override
  void dispose() {}
}

void runReducers(Iterable<SocketChannelReducer> reducers, StringConversion msg,
    APIState state, Function(APIState) changeState) {
  reducers.forEach(
    (reducer) => reducer.messageReducer(state, msg).map(
          identity: () {},
          connected: () => changeState(apiConnected),
        ),
  );
}
