import 'package:bird/signal_wave.dart';
import 'package:bird/smart_api.dart';
import 'package:dartz/dartz.dart' hide State;
import 'package:meta/meta.dart';
import 'package:pedantic/pedantic.dart';

/// Represents a traditional socket channel.
///
/// It uses a [SocketEndpointImpl] and interacts with its raw socket in a way
/// that respects the socket state and provides a different state for itself.
class SocketChannelServiceEndpoint<T> extends ServiceEndpoint<T>
    implements ReducerStateInterface {
  List<SocketChannelReducer> get reducers => channel.reducers;

  final SocketChannelPreferences channel;
  final bool ownsSocket;
  final SocketEndpointInterface Function(SocketChannelUrlProvider url) socket;
  final SocketChannelUrlProvider channelUrl;
  final T Function(StringConversion) converter;
  final IsConnectedPolicy whenIsConnectedPolicy;

  final Signal<APIState> _lastState = Signal(apiInitialized);
  SocketEndpointInterface _tempSocket;
  Disposable _socketSubscription;
  Disposable _messageSubscription;

  @protected
  Signal<Option<Either<AllErrors, T>>> _lastData;

  SocketChannelServiceEndpoint({
    /// Whether to dispose the socket that came from [socket] when this is disposed.
    @required this.ownsSocket,

    /// Url will be null if no [channelUrl] is given
    @required this.socket,
    @required this.channel,
    @required this.converter,
    this.channelUrl,
    this.whenIsConnectedPolicy = const ConnectImmediatelyPolicy(),
  }) {
    channel.reducers.forEach((a) => a.register(this));
    onNotConnectedAnymore();
    _lastData = Signal(
      none(),
      onListen: onConnected,
      onCancel: () {
        final newValue = addInitWhenActive(_lastState.value);
        if (newValue != null) _lastState.add(newValue);
        onNotConnectedAnymore();
      },
    );
  }

  /// This being a getter ensures that the synchronously available value is up-to-date.
  @override
  OEWave<T> get lastData => _lastData.wave.as(OEWave.from);

  /// This being a getter ensures that the synchronously available value is up-to-date.
  @override
  Wave<APIState> get lastState => _lastState.wave.distinct();

  /// Meant to only be used by reducers.
  @override
  void sendMessage(String msg) {
    if (lastState.value is APIStateConnected) _tempSocket.sendMessage(msg);
  }

  void onConnected() {
    if (_lastData.hasSubscribers && !_isDisposed) {
      if (_lastState.value is! APIStateInitialized) {
        _lastState.add(apiInitialized);
      }

      _lastState.add(apiLoading);

      _tempSocket = socket(channelUrl);

      _socketSubscription = _tempSocket.lastData.subscribe((data) async {
        if (!_isDisposed) {
          if (whenIsConnectedPolicy is ConnectImmediatelyPolicy &&
              _lastState.value is! APIStateConnected) {
            _lastState.add(apiConnected);
          }

          data.fold(() {}, (Either<AllErrors, SocketInterface> data) {
            data.fold(disconnect, (SocketInterface rawSocket) {
              channel.onReceivedSocket.fold(() {}, (passDataToSocket) {
                passDataToSocket(rawSocket.add);
              });

              _messageSubscription?.cancel();
              _messageSubscription =
                  _tempSocket.socketData.subscribe((optionalData) {
                optionalData.fold(() {
                  /// Do nothing.
                }, (str) {
                  if (whenIsConnectedPolicy
                          is IsConnectedOnFirstMessagePolicy &&
                      _lastState.value is! APIStateConnected) {
                    _lastState.add(apiConnected);
                  }

                  runReducers(reducers, str, _lastState.value, _lastState.add);
                  _lastData.add(Some(Right(converter(str))));
                });
              });
            });
          });
        }
      });
    }
  }

  void onNotConnectedAnymore() {
    channel.onLostSocket?.fold(() {}, (doThis) {
      if (_tempSocket != null) {
        (_tempSocket.lastData.value |
                const Left<AllErrors, SocketInterface>(null))
            .fold((error) {}, (socket) {
          doThis(socket.add);
        });
      }
    });
    _socketSubscription?.cancel();
    _socketSubscription = null;
    _messageSubscription?.cancel();
  }

  @override
  void reconnect() {
    if (lastState.value is APIStateDisconnected) onConnected();
  }

  @override
  void disconnect(AllErrors error) {
    if (_lastState.value is! APIStateDisconnected) {
      _lastData.add(Some(Left(error)));
      _lastState.add(apiDisconnected(error));
    }
    onNotConnectedAnymore();
  }

  bool _isDisposed = false;

  @override
  Future<void> dispose() async {
    _isDisposed = true;
    onNotConnectedAnymore();
    reducers.forEach((a) => a.dispose());
    if (_lastState.value is! APIStateDisconnected) {
      _lastState.add(apiDisconnected(cancelledError));
    }
    _lastState.close();
    _lastData.close();
    if (ownsSocket) unawaited(_tempSocket?.dispose());
  }
}
