import 'dart:async';

import 'package:bird/smart_api.dart';
import 'package:dartz/dartz.dart' hide State;
import 'package:dartz/dartz.dart' as z;
import 'package:dartz/dartz.dart';
import 'package:meta/meta.dart';

abstract class WebSocketProviderService {
  const WebSocketProviderService();

  SocketInterface createSocketForUrl(String url);
}

/// An interface for a socket connection.
abstract class SocketInterface {
  /// This will provide the raw data coming from the websocket connection.
  ///
  /// Errors or a None indicate that the socket was closed.
  OEWave<dynamic> get stream;

  /// Call this to send data through the socket connection to the server.
  void add(String str);

  /// Closes the socket which should call the 'onDone' event on [stream].
  Future<void> close();
}

/// Describes a websocket API.
///
/// There are secure and insecure websockets.
/// ws are insecure websockets (scheme: ws)
/// wss are secure websockets (scheme: wss).
///
/// All websocket definitions have to be _secure_ websocket connections.
///
/// Look through existing implementations to find how to implement a new description.
abstract class APIWebSocketDescription implements SocketChannelUrlProvider {
  /// Defines the host of an API path.
  ///
  /// Example: api.example.com
  final String host;

  /// Defines the port of an API path.
  ///
  /// Example: 9443.
  ///
  /// The default port for a secure websocket connection is 443 and
  /// should be explicitly set.
  final int port;

  /// Path of the websocket uri.
  ///
  /// Example: api/2/ws
  ///
  /// Can be null if not needed.
  final String path;

  const APIWebSocketDescription({
    @required this.host,
    this.port = 443,
    this.path,
  });

  /// Default url for [single] channels.
  ///
  /// Override to provide a default urlBuilder for [single].
  String defaultSingleUrlBuilder(
          String channelName, String id, APIWebSocketDescription description) =>
      description.buildUriFor();

  /// Default [SocketChannelPreferences] for [single] channels without custom preferences.
  ///
  /// Override to provide default [SocketChannelPreferences] for [single].
  SocketChannelPreferences defaultChannelPreferences(
          String name, String defid) =>
      SocketChannelPreferences(
        whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
      );

  /// Creates a single socket channel.
  ///
  /// Override [defaultSingleUrlBuilder] to create a custom url or pass one as [urlBuilder].
  /// The sockets definition url is the default.
  SingleChannelSocketDefinition<T> single<T>(
    String channelName, {
    @required
        T Function(StringConversion, SocketChannelURLAndMoreProvider) converter,
    SocketChannelPreferences Function(String channelName, String id)
        preferences,
    String Function(String) idBuilder,
    String Function(
            String channelName, String id, APIWebSocketDescription description)
        urlBuilder,
  }) {
    final channelID = (idBuilder ?? z.id)(channelName);
    return SingleChannelSocketDefinition(
      description: this,
      channelName: channelName,
      id: channelID,
      buildUrl: urlBuilder ?? defaultSingleUrlBuilder,
      preferences:
          (preferences ?? defaultChannelPreferences)(channelName, channelID),
      converter: converter,
    );
  }

  /// The URI for the websocket is built.
  ///
  /// The additional path is appended to [path] like
  /// [path]/[additionalPath].
  ///
  /// [args] are a map of either a String or a list of strings.
  /// These are the parameters you normally see in a url like
  ///
  /// ?abc=def&x=y,z
  String buildUriFor({
    String additionalPath,
    Map<String, dynamic /*String|Iterable<String>*/ > args,
  }) =>
      Uri(
              scheme: "wss",
              host: host,
              port: port,
              queryParameters: args,
              path: [path, additionalPath]
                  .where((a) => a != null)
                  .map((String s) => "/$s") // adds a leading slash to the front
                  .join(""))
          .toString();

  /// This socket itself is a [SocketChannelUrlProvider]. This allows the definition to be passed to
  /// the sockets instantiation for sockets that only take the root url and handle the rest differently.
  @override
  String get url => buildUriFor();
}

/// Don't store a single model for every usage.
/// Always instantiate a new one because reducers can be disposed.
class SocketChannelPreferences {
  final Option<void Function(Function(String) add)> onReceivedSocket;

  final Option<void Function(Function(String) add)> onLostSocket;

  final List<SocketChannelReducer> reducers;

  final IsConnectedPolicy whenIsConnectedPolicy;

  SocketChannelPreferences({
    @required this.whenIsConnectedPolicy,
    this.reducers = const [],
    String sub,
    String unsub,
  })  : this.onReceivedSocket =
            optionOf(sub != null ? ((add) => add(sub)) : null),
        this.onLostSocket =
            optionOf(unsub != null ? ((add) => add(unsub)) : null);
}

/// Each channel definition needs a url
abstract class SocketChannelUrlProvider {
  String get url;
}

/// Each channel definition needs a url
abstract class SocketChannelURLAndMoreProvider
    implements SocketChannelUrlProvider {
  String get name;

  String get id;

  @override
  String get url;
}

/// Used to define a single channels properties.
class SingleChannelSocketDefinition<T>
    implements SocketChannelURLAndMoreProvider {
  /// The unique component of a channel id without any other parameters added to it.
  ///
  /// Usually a channel will have a unique name that will be combined with other
  /// properties to get the full channel identifier.
  ///
  /// That dynamic property should be requested during the creation of this channel like this
  ///
  ///     SingleChannelSocketDefinition createChannel(Symbol symbol) => ...
  ///
  /// and combined with the channelName inside [buildId]
  final String channelName;

  /// Tells to which socket definition this belongs to which provides the url schema for
  /// the socket connection.
  final APIWebSocketDescription description;

  /// Combine the [channelName], the [id] and this channels [description] to a url as described
  /// in [SocketChannelDefinition.url].
  final String Function(
          String channelName, String id, APIWebSocketDescription description)
      buildUrl;

  final SocketChannelPreferences preferences;

  final T Function(StringConversion, SocketChannelURLAndMoreProvider) converter;

  T thisConverter(StringConversion conv) => converter(conv, this);

  const SingleChannelSocketDefinition({
    @required this.description,
    @required this.channelName,
    @required this.buildUrl,
    @required this.preferences,
    @required this.converter,
    @required this.id,
  });

  @override
  String get url => buildUrl(channelName, id, description);

  /// An id consists of the [channelName] and a dynamic property.
  ///
  /// This may seem redundant as this could be the channelName but some sockets
  /// become much easier to build when there is a distinction between its name and
  /// its id.
  @override
  final String id;

  @override
  String get name => channelName;
}
