import 'package:bird/smart_api.dart';
import 'package:dartz/dartz.dart';
import 'package:meta/meta.dart';

const List<APIState> apiInitLoadConnect = [
  apiInitialized,
  apiLoading,
  apiConnected
];

/// Describes the state of the current "connection"
abstract class APIState {
  const APIState();

  bool get isLoading => this is APIStateLoading;

  Option<AllErrors> get hasError {
    final self = this;
    if (self is APIStateDisconnected) {
      return self.error;
    } else {
      return none();
    }
  }

  // ignore: missing_return
  T visit<T>({
    @required T Function(APIStateInitialized state) initialized,
    @required T Function(APIStateConnected state) connected,
    @required T Function(APIStateLoading state) loading,
    @required T Function(APIStateDisconnected state) disconnected,
  }) {
    if (this is APIStateInitialized) {
      return initialized(this as APIStateInitialized);
    } else if (this is APIStateConnected) {
      return connected(this as APIStateConnected);
    } else if (this is APIStateLoading) {
      return loading(this as APIStateLoading);
    } else if (this is APIStateDisconnected) {
      return disconnected(this as APIStateDisconnected);
    }

    throw UnimplementedError("Missing implementation for ${this.runtimeType}");
  }
}

const APIState apiInitialized = APIStateInitialized();

/// Initial state
class APIStateInitialized extends APIState {
  const APIStateInitialized();

  @override
  bool operator ==(Object other) =>
      identical(this, other) || other is APIStateInitialized;

  @override
  int get hashCode => 0;
}

const APIState apiConnected = APIStateConnected();

/// A connection was successfully established.
class APIStateConnected extends APIState {
  const APIStateConnected();

  @override
  bool operator ==(Object other) =>
      identical(this, other) || other is APIStateConnected;

  @override
  int get hashCode => 0;
}

const APIState apiLoading = APIStateLoading();

/// Tells that new data is about be received.
///
/// REST: when data is being fetched.
/// Socket: while a connection has not been acknowledged yet.
class APIStateLoading extends APIState {
  const APIStateLoading();

  @override
  bool operator ==(Object other) =>
      identical(this, other) || other is APIStateLoading;

  @override
  int get hashCode => 0;
}

APIStateDisconnected apiDisconnected(AllErrors err) =>
    APIStateDisconnected(Some(err));

/// The user has to manually reconnect. This occurs when for example too many reconnect requests have been made.
///
/// REST: The user should not automatically be able to retry connecting to a server
///       indefinitely. This tells that a threshold has been crossed.
/// Socket: Same as REST.
class APIStateDisconnected extends APIState {
  final Option<AllErrors> error;

  const APIStateDisconnected(this.error);

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
      other is APIStateDisconnected &&
          runtimeType == other.runtimeType &&
          error == other.error;

  @override
  int get hashCode => error.hashCode;

  @override
  String toString() => '$APIStateDisconnected{error: $error}';
}
