import 'package:bird/smart_api.dart';
import 'package:dartz/dartz.dart';
import 'package:test/test.dart';

void main() {
    group("$AllErrors", () {
        test("disconnected", () {
            expect(apiDisconnected(const FailedParsingStorageError("a")),
                const APIStateDisconnected(Some(FailedParsingStorageError("a"))));
        });

        group("== && hashcode", () {
            group("$StorageErrors subclasses", () {
                test("$FailedParsingStorageError", () {
                    expect(const FailedParsingStorageError("a").asLeft<String>(),
                        const Left<dynamic, dynamic>(FailedParsingStorageError("b")));
                    expect(const FailedParsingStorageError("a"), const FailedParsingStorageError("b"));
                    expect(const FailedParsingStorageError("a"), const FailedParsingStorageError("a"));
                    expect(const FailedParsingStorageError("a").hashCode, const FailedParsingStorageError("a").hashCode);
                    expect(const FailedParsingStorageError("a").hashCode, const FailedParsingStorageError("b").hashCode);
                });

                test("$DataNotFoundStorageError", () {
                    // ignore: prefer_const_constructors
                    expect(notFoundInStorageError, DataNotFoundStorageError());
                    expect(const DataNotFoundStorageError(), const DataNotFoundStorageError());
                    expect(const DataNotFoundStorageError().hashCode, const DataNotFoundStorageError().hashCode);
                });

                test("$UnknownStorageError", () {
                    expect(unknownStorageError, UnknownStorageError(none(), none()));
                    expect(
                        unknownHintedStorageError(null, stacktrace: StackTrace.fromString("a")), unknownHintedStorageError(null));
                    expect(unknownHintedStorageError("a"), unknownStorageError);
                    expect(unknownStorageError.hashCode, unknownStorageError.hashCode);
                    expect(
                        unknownHintedStorageError("a", stacktrace: StackTrace.fromString("b")).hashCode,
                        unknownHintedStorageError("b", stacktrace: StackTrace.fromString("b")).hashCode,
                    );
                });
            });

            group("$APIErrors subclasses", () {
                test("$CancelledAPIError", () {
                    expect(cancelledError, const CancelledAPIError());
                    expect(cancelledError.left<String>(), const Left<dynamic, dynamic>(cancelledError));
                    expect(const CancelledAPIError(), const CancelledAPIError());
                    // ignore: prefer_const_constructors
                    expect(const CancelledAPIError(), CancelledAPIError());
                    expect(const CancelledAPIError(), isNot(unknownStorageError));
                    expect(const CancelledAPIError().hashCode, const CancelledAPIError().hashCode);
                });

                test("$LocalNotEnoughCreditsAvailableAPIError", () {
                    expect(noCreditsError, const LocalNotEnoughCreditsAvailableAPIError());
                    // ignore: prefer_const_constructors
                    expect(const LocalNotEnoughCreditsAvailableAPIError(), LocalNotEnoughCreditsAvailableAPIError());
                    // ignore: prefer_const_constructors
                    expect(const LocalNotEnoughCreditsAvailableAPIError(), isNot(CancelledAPIError()));
                    expect(const LocalNotEnoughCreditsAvailableAPIError(), const LocalNotEnoughCreditsAvailableAPIError());
                    expect(const LocalNotEnoughCreditsAvailableAPIError().hashCode,
                        const LocalNotEnoughCreditsAvailableAPIError().hashCode);
                });

                test("$APIStatusError", () {
                    expect(statusError(200), statusError(200));
                    expect(statusError(201), isNot(statusError(200)));
                    expect(statusError(200).hashCode, statusError(200).hashCode);
                    expect(statusError(200).toString(), contains("$APIStatusError"));
                });

                test("$NoInternetAPIError", () {
                    expect(noInternetError, const NoInternetAPIError());
                    // ignore: prefer_const_constructors
                    expect(const NoInternetAPIError(), NoInternetAPIError());
                    expect(const NoInternetAPIError(), isNot(unknownStorageError));
                    expect(const NoInternetAPIError().hashCode, const NoInternetAPIError().hashCode);
                });

                test("$TimeoutAPIError", () {
                    // ignore: prefer_const_constructors
                    expect(timeoutError, TimeoutAPIError());
                    expect(const TimeoutAPIError(), const TimeoutAPIError());
                    expect(const TimeoutAPIError(), isNot(unknownStorageError));
                    expect(const TimeoutAPIError().hashCode, const TimeoutAPIError().hashCode);
                });

                test("$APIUnexpectedResponseError", () {
                    expect(unexpectedResponseError(const ResponseInformation.none()),
                        const APIUnexpectedResponseError(ResponseInformation.none()));
                    expect(const APIUnexpectedResponseError(null), const APIUnexpectedResponseError(null));
                    // ignore: prefer_const_constructors
                    expect(const APIUnexpectedResponseError(ResponseInformation.none()),
                        const APIUnexpectedResponseError(ResponseInformation.none()));
                    expect(const APIUnexpectedResponseError(null).hashCode, const APIUnexpectedResponseError(null).hashCode);
                    expect(const APIUnexpectedResponseError(null).hashCode, const APIUnexpectedResponseError(null).hashCode);
                });

                test("$UnknownAPIError", () {
                    expect(unknownHintedAPIError("a", stacktrace: null), unknownHintedAPIError("a", stacktrace: null));
                    expect(unknownHintedAPIError("a", stacktrace: null), unknownHintedAPIError("b", stacktrace: null));
                    expect(unknownHintedAPIError("a", stacktrace: null).hashCode,
                        unknownHintedAPIError("a", stacktrace: null).hashCode);
                    expect(unknownHintedAPIError("a", stacktrace: null).hashCode,
                        unknownHintedAPIError("b", stacktrace: null).hashCode);
                });

                test("$APIResponseCustomErrorDetectedAPIError", () {
                    // ignore: prefer_const_constructors
                    expect(vendorError, APIResponseCustomErrorDetectedAPIError());
                    expect(const APIResponseCustomErrorDetectedAPIError(), const APIResponseCustomErrorDetectedAPIError());
                    expect(const APIResponseCustomErrorDetectedAPIError().hashCode,
                        const APIResponseCustomErrorDetectedAPIError().hashCode);
                });
            });
        });
        group("map AllDataErrors", () {
            test("Storage errors", () {
                expect(unknownAPIError.map((a) => a, (a) => null), unknownAPIError);
                expect(unknownStorageError.map((a) => null, (a) => a), unknownStorageError);
            });

            test("API errors", () {
                void testMapError(AllErrors error) => expect(error.map(id, id), error);
                testMapError(const CancelledAPIError());
                testMapError(const LocalNotEnoughCreditsAvailableAPIError());
                testMapError(const NoInternetAPIError());
                testMapError(statusError(404));
                testMapError(const TimeoutAPIError());
                testMapError(const APIUnexpectedResponseError(null));
                testMapError(unknownAPIError);
                expect(unknownHintedAPIError("a", stacktrace: StackTrace.fromString("a")).toString(),
                    'UnknownAPIError{description: Some(a), stacktrace: Some(a)');
                testMapError(const APIResponseCustomErrorDetectedAPIError());
            });
        });

        test("map DataStorageErrors", () {
            void testMapError(StorageErrors error) => expect(error.subMap(id, id, id), error);
            testMapError(unknownStorageError);
            testMapError(const DataNotFoundStorageError());
            testMapError(const FailedParsingStorageError("description"));
        });

        test("map APIErrors", () {
            void testMapError(APIErrors error) =>
                expect(error.subMap(
                    id,
                    id,
                    id,
                    id,
                    id,
                    id,
                    id,
                    id), error);
            testMapError(const CancelledAPIError());
            testMapError(const LocalNotEnoughCreditsAvailableAPIError());
            testMapError(const NoInternetAPIError());
            testMapError(statusError(404));
            testMapError(const TimeoutAPIError());
            testMapError(const APIUnexpectedResponseError(null));
            expect(const APIUnexpectedResponseError(null).toString(), const TypeMatcher<String>());
            testMapError(unknownAPIError);
            testMapError(vendorError);
        });
    });
}