import 'dart:async';

import 'package:bird/dart_filter.dart';
import 'package:bird/smart_api.dart';
import 'package:bird/smart_json.dart';
import 'package:dartz/dartz.dart';
import 'package:mockito/mockito.dart';
import 'package:test/test.dart';

import '../mock.dart';

void main() {
    const TestRestAPI api = TestRestAPI();

    group("$APIRestDescription", () {
        test("host", () {
            expect(api.host, "example.com");
        });

        test("path", () {
            expect(api.path, "api");
        });

        test("get minimal", () {
            final minimal = api.exampleGetMinimal;
            expect(minimal.arguments.httpMethod, "GET");
            expect(minimal.arguments.postBody, null);
            expect(minimal.api, api);
            expect(minimal.path, null);
            expect(minimal.cmd, null);
            expect(minimal.expectedContentType, <FilterCriteria<String>>[]);
            expect(minimal.uniqueIdentifier, "example.com/GET/api");
        });

        test("get full", () {
            final full = api.exampleGetFull;

            /// https://passwordsgenerator.net/md5-hash-generator/
            /// args: a=b&c=d&c=e => C4BFCA19959708FA417FC6548A4A00D2
            expect(full.uniqueIdentifier, "example.com/GET/api/create/CREATE/c4bfca19959708fa417fc6548a4a00d2");
            expect(full.arguments.headers(full), {"header-key": "header-value"});
            expect(full.arguments.postBody, null);
            expect(full.expectedContentType.length, 1);
            expect(full.expectedContentType.first, APIContentTypes.anyJson);
        });

        test("post minimal", () {
            final minimal = api.examplePostMinimal;
            expect(minimal.arguments.httpMethod, "POST");
            expect(minimal.api, api);
            expect(minimal.path, null);
            expect(minimal.cmd, null);
            expect(minimal.expectedContentType, <FilterCriteria<String>>[]);
            expect(minimal.uniqueIdentifier, "example.com/POST/api");
        });

        test("post full", () {
            final full = api.examplePostFull;

            /// https://passwordsgenerator.net/md5-hash-generator/
            /// postBody: {} => 99914B932BD37A50B983C5E7C90AE93B
            expect(full.uniqueIdentifier, "example.com/POST/api/create/CREATE/99914b932bd37a50b983c5e7c90ae93b");
            expect(full.arguments.headers(full), {"header-key": "header-value"});
            expect(full.expectedContentType.length, 1);
            expect(full.expectedContentType.first, APIContentTypes.anyJson);
        });

        test("buildUri()", () {
            expect(api.exampleGetMinimal.buildUri().toString(), "https://example.com/api");
            expect(api.exampleGetFull.buildUri().toString(), "https://example.com/api/create?a=b&c=d&c=e");
            expect(api.examplePostMinimal.buildUri().toString(), "https://example.com/api");
            expect(() => api.exampleGetInvalidArguments.buildUri(), throwsA(const TypeMatcher<ArgumentError>()));
        });

        test("buildUri() takes int parameters", () {
            final desc = APIRestMethodDescription(
                api: api,
                arguments: APIHTTPMethod.get(
                    getParam: (_) => {"command": "a", "period": 3},
                ),
            );
            expect(desc.buildUri().toString(), "https://example.com/api?command=a&period=3");
        });
    });

    group("$APIRestMethodDescription", () {
        test("null checks", () {
            expect(() => APIRestMethodDescription(api: api, arguments: null), throwsA(const TypeMatcher<AssertionError>()));
            // ignore: prefer_const_constructors
            expect(() => APIRestMethodDescription(api: null, arguments: const APIHTTPMethod.get()),
                throwsA(const TypeMatcher<AssertionError>()));
            expect(() => APIRestMethodDescription(api: api, arguments: const APIHTTPMethod.get(), expectedContentType: null),
                throwsA(const TypeMatcher<AssertionError>()));
        });

        group("methodFetch", () {
            test("return what service returns", () async {
                final WebRequestService service = MockWebRequestService(() => const Right(""));
                final DataResponse<String> response = await api.exampleGetMinimal.methodFetch(service);
                expect(response.response, rightAPIErrorsDataIs(""));
            });
        });
    });

    group("$APIContentTypes", () {
        test("$APIContentTypes.anyJson", () {
            expect(APIContentTypes.anyJson.accepts("application/json"), true);
            expect(APIContentTypes.anyJson.accepts("application/json utf8"), true);
            expect(APIContentTypes.anyJson.accepts("application/rss+xml"), false);
        });

        test("$APIContentTypes.anyRssXml", () {
            expect(APIContentTypes.anyRssXml.accepts("application/rss+xml"), true);
            expect(APIContentTypes.anyRssXml.accepts("application/rss+xml utf8"), true);
            expect(APIContentTypes.anyRssXml.accepts("application/json"), false);
        });

        test("$APIContentTypes.textXml", () {
            expect(APIContentTypes.textXml.accepts("text/xml"), true);
            expect(APIContentTypes.textXml.accepts("text/xml"), true);
            expect(APIContentTypes.textXml.accepts("application/json"), false);
        });
    });

    group("$APIHTTPMethod", () {
        group("GET", () {
            test("get arguments not null", () {
                // ignore: prefer_const_constructors
                expect(() => APIHTTPMethod.get(headers: null), throwsA(const TypeMatcher<AssertionError>()));
            });

            test("headers initialized to {}", () {
                expect(const APIHTTPMethod.get().headers(api.get()), <String, String>{});
            });

            test("get is GET", () {
                expect(const APIHTTPMethod.get().httpMethod, "GET");
            });

            test("lowFetch, data is returned", () async {
                final WebRequestService service = MockWebRequestService(() => const Right("{}"));
                final DataResponse<String> response = await const APIHTTPMethod.get().lowFetch(
                    const APIRestMethodDescription(api: TestRestAPI(), arguments: APIHTTPMethod.get()), service);
                expect(response.response, rightAPIErrorsDataIs("{}"));
            });

//            test("lowFetch, verify get is called", () async {
//                final Dio mockDio = MockitoDio();
//                await const APIHTTPMethod.get().lowFetch(const APIRestMethodDescription(api: TestRestAPI(), arguments: APIHTTPMethod.get()), null, mockDio);
//                verify(mockDio.request<dynamic>(
//                    argThat(equals("https://example.com/api")),
//                    data: argThat(isNull, named: "data"),
//                    options: argThat(const TypeMatcher<Options>().having((a) => a.method, "", equals("GET")), named: "options"),
//                    cancelToken: argThat(isNull, named: "cancelToken"),
//                    onUploadProgress: argThat(isNull, named: "onUploadProgress"),
//                ));
//            });
        });
        group("POST", () {
            test("post arguments not null", () {
                // ignore: prefer_const_constructors
                expect(() => APIHTTPMethod.post(headers: null), throwsA(const TypeMatcher<AssertionError>()));
            });

            test("headers initialized to {}", () {
                expect(const APIHTTPMethod.post().headers(api.post()), <String, String>{});
            });

            test("post is POST", () {
                expect(const APIHTTPMethod.post().httpMethod, "POST");
            });

//            test("lowFetch, data is returned", () async {
//                final Dio mockDio = MockDio(() => Response<String>(data: "{}", statusCode: 200));
//                final DataResponse<String> response = await const APIHTTPMethod.post().lowFetch(
//                    const APIRestMethodDescription(api: TestRestAPI(), arguments: APIHTTPMethod.post()), null, mockDio);
//                expect(response.response, rightAPIErrorsDataIs("{}"));
//            });
//
//            test("lowFetch, verify post is called", () async {
//                final Dio mockDio = MockitoDio();
//                await const APIHTTPMethod.post().lowFetch(const APIRestMethodDescription(api: TestRestAPI(), arguments: APIHTTPMethod.post()), null, mockDio);
//                verify(mockDio.request<dynamic>(
//                    argThat(equals("https://example.com/api")),
//                    data: argThat(isNull, named: "data"),
//                    options: argThat(const TypeMatcher<Options>().having((a) => a.method, "", equals("POST")), named: "options"),
//                    cancelToken: argThat(isNull, named: "cancelToken"),
//                    onUploadProgress: argThat(isNull, named: "onUploadProgress"),
//                ));
//            });
        });
    });

    group("$DataResponse", () {
        test("args not null", () {
            expect(() => DataResponse<Object>(started: null, ended: DateTime.now(), response: const Left(null)),
                throwsA(const TypeMatcher<AssertionError>()));
            expect(() => DataResponse<Object>(started: DateTime.now(), ended: null, response: const Left(null)),
                throwsA(const TypeMatcher<AssertionError>()));
            expect(() => DataResponse<Object>(started: DateTime.now(), ended: DateTime.now(), response: null),
                throwsA(const TypeMatcher<AssertionError>()));
        });

        test("started before ended", () {
            expect(DataResponse<Object>(started: DateTime(2000), ended: DateTime(2001), response: const Left(null)),
                const TypeMatcher<DataResponse>());
            expect(DataResponse<Object>(started: DateTime(2000), ended: DateTime(2000), response: const Left(null)),
                const TypeMatcher<DataResponse>());
            expect(() => DataResponse<Object>(started: DateTime(2000), ended: DateTime(1999), response: const Left(null)),
                throwsA(const TypeMatcher<AssertionError>()));
        });

        test("constructor.measure", () async {
            final a = await DataResponse.measure(() async {
                await Future<void>.delayed(const Duration(milliseconds: 10));
                return const Right(null);
            });
            expect(a, const TypeMatcher<DataResponse>());
            expect(a.started.millisecondsSinceEpoch <= a.ended.millisecondsSinceEpoch, true);
            expect(a.response, const TypeMatcher<Right>());
        });

        test("toString", () async {
            final a = DataResponse<Object>(started: DateTime.now(), ended: DateTime.now(), response: const Left(unknownAPIError));
            expect(a.toString(), contains("$DataResponse"));
        });
    });

    group("$RAWModelUpdate", () {
        /// Nothing to test
    });

    group("$ModelUpdate", () {
        test("requestOnlyT Left", () async {
            final _MockModelUpdate value = _MockModelUpdate(const Left(unknownAPIError));
            expect(await value.requestOnlyT(), const TypeMatcher<Left>());
        });

        test("requestOnlyT Right", () async {
            final _MockModelUpdate value = _MockModelUpdate(Right(Tuple2(TestModel()
                ..b.safeSet("bla"), (TestModel()
                ..b.safeSet("bla")).convert().to())));
            expect(await value.requestOnlyT(), Right<AllErrors, TestModel>(TestModel()
                ..b.safeSet("bla")));
        });

        test("request Left", () async {
            final _MockModelUpdate value = _MockModelUpdate(const Left(unknownAPIError));
            expect(await value.request(), const TypeMatcher<Left>());
        });

        test("request Right", () async {
            final _MockModelUpdate value = _MockModelUpdate(Right(Tuple2(TestModel()
                ..b.safeSet("bla"), (TestModel()
                ..b.safeSet("bla")).convert().to())));
            expect(await value.request(), Right<AllErrors, Tuple2<TestModel, String>>(Tuple2(
                TestModel()
                    ..b.safeSet("bla"),
                (TestModel()
                    ..b.safeSet("bla")).convert().to(),
            )));
        });
    });

    group("$OneModelUpdate", () {
        test("arguments not null", () {
            expect(() =>
                OneModelUpdate<TestModel>(method: null, tCreator: () => TestModel(), requestService: MockWebRequestService(null)),
                throwsA(const TypeMatcher<AssertionError>()));
            expect(() => OneModelUpdate<TestModel>(method: api.m, tCreator: null, requestService: MockWebRequestService(null)),
                throwsA(const TypeMatcher<AssertionError>()));
            expect(() => OneModelUpdate<TestModel>(method: api.m, tCreator: () => TestModel(), requestService: null),
                throwsA(const TypeMatcher<AssertionError>()));
        });

        group("request", () {
            test("fails parsing model and resets to default", () async {
                const String invalidJson = '{"a":"b';
                final OneModelUpdate oneModel = OneModelUpdate<TestModel>(
                    method: api.m, tCreator: () => TestModel(),
                    requestService: MockWebRequestService(() => const Right(invalidJson)),
                );
                expect(
                    await oneModel.request(),
                    leftAPIErrorIsWithRight<UnknownAPIError, Tuple2<TestModel, String>>(),
                );
            });

            test("parses model successfully", () async {
                const String json = '{"a":"b", "b":"a"}';
                final OneModelUpdate oneModel = OneModelUpdate<TestModel>(
                    method: api.m, tCreator: () => TestModel(),
                    requestService: MockWebRequestService(() => const Right(json)),
                );
                expect(
                    await oneModel.request(),
                    rightAPIErrorsDataIs(Tuple2(TestModel()
                        ..a.safeSet("b")
                        ..b.safeSet("a"), json)),
                );
            });

            test("catches custom errors", () async {
                final OneModelUpdate oneModel = OneModelUpdate<TestModel>(
                    method: api.m, tCreator: () => TestModel(),
                    requestService: MockWebRequestService(() => Left(statusError(201))),
                );
                expect(
                    await oneModel.request(),
                    leftAPIErrorIsWithRight<APIStatusError, Tuple2<TestModel, String>>(),
                );
            });
        });
    });

    group("$RawOneModelUpdate", () {
        test("arguments not null", () {
            expect(() => RawOneModelUpdate<String>(method: null, tCreator: (a) => a, requestService: MockWebRequestService(null)),
                throwsA(const TypeMatcher<AssertionError>()));
            expect(() => RawOneModelUpdate<String>(method: api.m, tCreator: null, requestService: MockWebRequestService(null)),
                throwsA(const TypeMatcher<AssertionError>()));
            expect(() => RawOneModelUpdate<String>(method: api.m, tCreator: (a) => a, requestService: null),
                throwsA(const TypeMatcher<AssertionError>()));
        });

        group("request", () {
            test("fails parsing model and resets to default", () async {
                const String invalidJson = '{"a":"b';
                final RawOneModelUpdate oneModel = RawOneModelUpdate<String>(
                    method: api.m, tCreator: (a) => a,
                    requestService: MockWebRequestService(() => const Right(invalidJson)),
                );
                expect(
                    await oneModel.request(),
                    rightAPIErrorsDataIs(Tuple2(invalidJson, invalidJson)),
                );
            });

            test("parses model successfully", () async {
                const String json = '{"a":"b", "b":"a"}';
                final RawOneModelUpdate oneModel = RawOneModelUpdate<String>(
                    method: api.m, tCreator: (a) => a,
                    requestService: MockWebRequestService(() => const Right(json)),
                );
                expect(
                    await oneModel.request(),
                    rightAPIErrorsDataIs(Tuple2(json, json)),
                );
            });

            test("caches custom errors", () async {
                final RawOneModelUpdate oneModel = RawOneModelUpdate<String>(
                    method: api.m, tCreator: (a) => a,
                    requestService: MockWebRequestService(() => Left(statusError(201))),
                );
                expect(
                    await oneModel.request(),
                    leftAPIErrorIsWithRight<APIStatusError, Tuple2<String, String>>(),
                );
            });
        });
    });

    group("$APISingleDataAccess", () {
        test("arguments not null", () {
//            final OneModelUpdate<TestModel> oneModel = OneModelUpdate<TestModel>(method: api.m, tCreator: () => TestModel(),
//                requestService: MockWebRequestService(null));
//            expect(() => APISingleDataAccess<TestModel>(
//                apiStorage: null,
//                modelUpdate: oneModel,
//            ), throwsSymbolAssert(#a));
            expect(() =>
                APISingleDataAccess<TestModel>(
//                    apiStorage: MockStorageAccess(
//                        getCacheVar: null,
//                        hasCacheVar: none(),
//                        putCacheVar: {},
//                    ),
                    modelUpdate: null,
                ),
                throwsSymbolAssert(#b),
            );
        });

        test("fetchNewData", () async {
            final model = APISingleDataAccess<TestModel2>(
                modelUpdate: OneModelUpdate<TestModel2>(
                    method: api.m, tCreator: () => TestModel2(),
                    requestService: MockWebRequestService(() => const Right('{"x":"3x","y":"3y"}')),
                ),
            );

            final newData = await model.request();
            final newDataModel = TestModel2()
                ..x.safeSet("3x")
                ..y.safeSet("3y");

            expect(newData, rightAllErrorsDataIs(Tuple2(newDataModel, '{"x":"3x","y":"3y"}')));

            final error = await model.request();

            /// One error recognized means error wrapper is present.
            expect(error, Right<AllErrors, Tuple2<TestModel2, String>>(Tuple2(newDataModel, newDataModel.toString())));
        });
    });

    group("$APIDoubleDataAccess", () {
        final _MockAPISingleDataAccess<TestModel> mockA = _MockAPISingleDataAccess();
        final _MockAPISingleDataAccess<TestModel2> mockB = _MockAPISingleDataAccess();

        test("arguments not null", () {
            expect(() =>
                APIDoubleDataAccess<TestModel, TestModel2, String>(freeModelUpdate: null, dependentModelUpdate: (a) => mockB),
                throwsSymbolAssert(#a));
            expect(() => APIDoubleDataAccess<TestModel, TestModel2, String>(freeModelUpdate: mockA, dependentModelUpdate: null),
                throwsSymbolAssert(#b));
        });

//        group("hasData", () {
//            test("free model retrieveFromCache is null ${#a}", () async {
//                reset(mockA);
//                reset(mockB);
//                expect(
//                        () =>
//                        APIDoubleDataAccess<TestModel, TestModel2, String>(
//                            freeModelUpdate: mockA,
//                            dependentModelUpdate: (a) => mockB,
//                        ).hasData(),
//                    throwsSymbolAssert(#a),
//                );
//            });
//
//            test("free model returns null at retrieveFromCache ${#b}", () async {
//                reset(mockA);
//                reset(mockB);
//                when(mockA.retrieveFromCache()).thenAnswer((a) async => null);
//                expect(
//                        () =>
//                        APIDoubleDataAccess<TestModel, TestModel2, String>(
//                            freeModelUpdate: mockA,
//                            dependentModelUpdate: (a) => mockB,
//                        ).hasData(),
//                    throwsSymbolAssert(#b),
//                );
//            });
//
//            test("free model no age ${#c}", () async {
//                reset(mockA);
//                reset(mockB);
//                when(mockA.retrieveFromCache()).thenAnswer((a) async => const CachedData(age: None(), data: Left(null)));
//                expect(
//                    await APIDoubleDataAccess<TestModel, TestModel2, String>(
//                        freeModelUpdate: mockA,
//                        dependentModelUpdate: (a) => mockB,
//                    ).hasData(),
//                    const None<DateTime>(),
//                );
//            });
//
//            test("free model age no data ${#d}", () async {
//                reset(mockA);
//                reset(mockB);
//                when(mockA.retrieveFromCache()).thenAnswer((a) async =>
//                    CachedData(age: Some(DateTime(2000)), data: const Left(null)));
//                expect(
//                    await APIDoubleDataAccess<TestModel, TestModel2, String>(
//                        freeModelUpdate: mockA,
//                        dependentModelUpdate: (a) => mockB,
//                    ).hasData(),
//                    const None<DateTime>(),
//                );
//            });
//
//            test("free model second model is null ${#e}", () async {
//                reset(mockA);
//                reset(mockB);
//                when(mockA.retrieveFromCache()).thenAnswer((a) async =>
//                    CachedData(age: Some(DateTime(2000)), data: right(TestModel())));
//                expect(
//                        () =>
//                        APIDoubleDataAccess<TestModel, TestModel2, String>(
//                            freeModelUpdate: mockA,
//                            dependentModelUpdate: (a) => null,
//                        ).hasData(),
//                    throwsSymbolAssert(#e),
//                );
//            });
//
//            test("successful return dependent data hasData ${#f}", () async {
//                reset(mockA);
//                reset(mockB);
//                when(mockA.retrieveFromCache()).thenAnswer((a) async =>
//                    CachedData(age: Some(DateTime(2000)), data: right(TestModel())));
//                when(mockB.hasData()).thenAnswer((a) async => Some(DateTime(2001)));
//                expect(
//                    await APIDoubleDataAccess<TestModel, TestModel2, String>(
//                        freeModelUpdate: mockA,
//                        dependentModelUpdate: (a) => mockB,
//                    ).hasData(),
//                    Some(DateTime(2001))
//                );
//            });
//        });

        group("fetchNewData", () {
            test("free model method not null ${#a}", () {
                reset(mockA);
                reset(mockB);
                expect(
                        () =>
                        APIDoubleDataAccess<TestModel, TestModel2, String>(
                            freeModelUpdate: mockA,
                            dependentModelUpdate: (a) => mockB,
                        ).requestOnlyT(),
                    throwsSymbolAssert(#a),
                );
            });

            test("free model returns null at fetchNewData ${#b}", () {
                reset(mockA);
                reset(mockB);
                when(mockA.requestOnlyT()).thenAnswer((a) async => null);
                expect(
                        () =>
                        APIDoubleDataAccess<TestModel, TestModel2, String>(
                            freeModelUpdate: mockA,
                            dependentModelUpdate: (a) => mockB,
                        ).requestOnlyT(),
                    throwsSymbolAssert(#b),
                );
            });

            test("dependent model returns null at fetchNewData ${#c}", () async {
                reset(mockA);
                reset(mockB);
                when(mockA.requestOnlyT()).thenAnswer((a) async => right(TestModel()));
                expect(() =>
                    APIDoubleDataAccess<TestModel, TestModel2, String>(
                        freeModelUpdate: mockA,
                        dependentModelUpdate: (a) => null,
                    ).requestOnlyT(),
                    throwsSymbolAssert(#c),
                );
            });

            test("successful fetch ${#d}", () async {
                reset(mockA);
                reset(mockB);
                when(mockA.requestOnlyT())
                    .thenAnswer((a) async => right(TestModel()));
                when(mockB.request())
                    .thenAnswer((a) async => right(Tuple2(TestModel2(), TestModel2().toString())));
                expect(await APIDoubleDataAccess<TestModel, TestModel2, String>(
                    freeModelUpdate: mockA,
                    dependentModelUpdate: (a) => mockB,
                ).request(),
                    rightAllErrorsDataIs(Tuple2(TestModel2(), TestModel2().toString())),
                );
            });
        });
    });

    group("$APIAccumulatedDataAccess", () {
        final _MockAPISingleDataAccess<TestModel2> mock2 = _MockAPISingleDataAccess();

        test("arguments not null", () async {
            reset(mock2);

            expect(() =>
                APIAccumulatedDataAccess<TestModel, TestModel2>(
                    source: null,
                    modelToUpdate: (TestModel first) => mock2,
                ),
                throwsSymbolAssert(#a),
            );

            expect(() =>
                APIAccumulatedDataAccess<TestModel, TestModel2>(
                    source: TestModel3().m.nullVal(),
                    modelToUpdate: null,
                ),
                throwsSymbolAssert(#b),
            );
        });

//        group("hasData", () {
//            test("is null ${#a}", () async {
//                reset(mock2);
//                expect(
//                        () =>
//                        APIAccumulatedDataAccess<TestModel, TestModel2>(
//                            source: TestModel3().m.nullVal(),
//                            modelToUpdate: (TestModel first) => null,
//                        ).hasData(),
//                    throwsSymbolAssert(#a),
//                );
//            });
//
//            test("returns null ${#b}", () async {
//                reset(mock2);
//                when(mock2.hasData()).thenAnswer((a) async => null);
//                expect(
//                        () =>
//                        APIAccumulatedDataAccess(
//                            source: TestModel3().m.nullVal(),
//                            modelToUpdate: (TestModel first) => mock2,
//                        ).hasData(),
//                    throwsSymbolAssert(#b),
//                );
//            });
//
//            test("if one is not present -> returns none ${#c}", () async {
//                reset(mock2);
//
//                when(mock2.hasData())
//                    ..thenAnswer((a) async => none());
//
//                expect(
//                    await APIAccumulatedDataAccess(
//                        source: TestModel3().m.nullVal(),
//                        modelToUpdate: (TestModel first) => mock2,
//                    ).hasData(),
//                    const TypeMatcher<None>(),
//                );
//
//                final _MockAPISingleDataAccess<TestModel2> mock22 = _MockAPISingleDataAccess();
//
//                when(mock22.hasData())
//                    ..thenAnswer((a) async => Some(DateTime(2000)));
//
//                expect(
//                    await APIAccumulatedDataAccess<APISingleDataAccess<TestModel2>, TestModel2>(
//                        source: [mock2, mock22],
//                        modelToUpdate: (a) => a,
//                    ).hasData(),
//                    const TypeMatcher<None>(),
//                );
//            });
//
//            test("empty source list ${#d}", () async {
//                reset(mock2);
//                expect(
//                    await APIAccumulatedDataAccess<TestModel, TestModel2>(
//                        source: [],
//                        modelToUpdate: (TestModel first) => mock2,
//                    ).hasData(),
//                    const TypeMatcher<None>(),
//                );
//            });
//
//            test("returns last ${#e}", () async {
//                reset(mock2);
//
//                final _MockAPISingleDataAccess<TestModel2> mock22 = _MockAPISingleDataAccess();
//
//                when(mock2.hasData())
//                    ..thenAnswer((a) async => Some(DateTime(1999)));
//
//                when(mock22.hasData())
//                    ..thenAnswer((a) async => Some(DateTime(2000)));
//
//                expect(
//                    await APIAccumulatedDataAccess<APISingleDataAccess<TestModel2>, TestModel2>(
//                        source: [mock2, mock22],
//                        modelToUpdate: (a) => a,
//                    ).hasData(),
//                    Some(DateTime(2000)),
//                );
//            });
//        });

        group("fetchNewData", () {
            test("model to update is not null ${#a}", () async {
                expect(
                        () =>
                        APIAccumulatedDataAccess<APISingleDataAccess<TestModel2>, TestModel2>(
                            source: [null],
                            modelToUpdate: (a) => null,
                        ).request(),
                    throwsSymbolAssert(#a),
                );
            });

            test("model to update does not return null ${#b}", () async {
                reset(mock2);
                when(mock2.requestOnlyT()).thenAnswer((a) async => null);
                expect(
                        () =>
                        APIAccumulatedDataAccess<APISingleDataAccess<TestModel2>, TestModel2>(
                            source: [mock2],
                            modelToUpdate: (a) => a,
                        ).requestOnlyT(),
                    throwsSymbolAssert(#b),
                );
            });

            test("one error means error is returned ${#c}", () async {
                reset(mock2);

                final _MockAPISingleDataAccess<TestModel2> mock22 = _MockAPISingleDataAccess();
                final _MockAPISingleDataAccess<TestModel2> mock222 = _MockAPISingleDataAccess();

                when(mock2.request()).thenAnswer((a) async => right(Tuple2(TestModel2(), TestModel2().toString())));

                when(mock22.request()).thenAnswer((a) async => left(unknownAPIError));

                when(mock222.request()).thenAnswer((a) async => right(Tuple2(TestModel2(), TestModel2().toString())));

                expect(
                    await APIAccumulatedDataAccess<APISingleDataAccess<TestModel2>, TestModel2>(
                        source: [mock2, mock22, mock222],
                        modelToUpdate: (a) => a,
                    ).request(),
                    leftAllErrorIsWith<UnknownAPIError, Tuple2<IList<TestModel2>, IList<String>>>(),
                );
            });
        });

        test("data was accumulated ${#d}", () async {
            reset(mock2);
            final _MockAPISingleDataAccess<TestModel2> mock22 = _MockAPISingleDataAccess();
            when(mock2.request()).thenAnswer((a) async => right(Tuple2(TestModel2(), TestModel2().toString())));
            when(mock22.request()).thenAnswer((a) async => right(Tuple2(TestModel2(), TestModel2().toString())));
            expect(
                await APIAccumulatedDataAccess<APISingleDataAccess<TestModel2>, TestModel2>(
                    source: [mock2, mock22],
                    modelToUpdate: (a) => a,
                ).request(),
                rightAllErrorsDataIs(Tuple2(
                    IList.from([TestModel2(), TestModel2()]), IList.from([TestModel2().toString(), TestModel2().toString()]))),
            );
        });

        test("first error is returned ${#e}", () async {
            reset(mock2);
            final _MockAPISingleDataAccess<TestModel2> mock22 = _MockAPISingleDataAccess();

            when(mock2.request()).thenAnswer((a) async => left(unknownAPIError));
            when(mock22.request()).thenAnswer((a) async => left(statusError(123)));
            expect(
                await APIAccumulatedDataAccess<APISingleDataAccess<TestModel2>, TestModel2>(
                    source: [mock2, mock22],
                    modelToUpdate: (a) => a,
                ).request(),
                leftAllErrorIsWith<UnknownAPIError, Tuple2<IList<TestModel2>, IList<String>>>(),
            );
        });
    });

    group("$APIState", () {
        /// Nothing to test
    });
}

class _MockModelUpdate extends RAWModelUpdate<TestModel, String> {

    final Either<AllErrors, Tuple2<TestModel, String>> ret;

    _MockModelUpdate(this.ret);

    @override
    Future<Either<AllErrors, Tuple2<TestModel, String>>> request() async => ret;
}

class _MockAPISingleDataAccess<T extends SmartJSON<T>> extends Mock implements APISingleDataAccess<T> {}
