import 'dart:async';

import 'package:bird/signal_wave.dart';
import 'package:bird/smart_api.dart';
import 'package:bird/smart_json.dart';
import 'package:dartz/dartz.dart';
import 'package:mockito/mockito.dart';
import 'package:test/test.dart';

import '../service_endpoint_spec.dart';

void main() {
    group("$RestServiceEndpoint", () {
        group("SPEC", () {
            _MockRAWModelUpdate okModel() {
                final model = _MockRAWModelUpdate();
                when(model.request())
                    .thenAnswer((_) async {
                    return Right(Tuple2(TestModel(), TestModel().convert().to()));
                });
                return model;
            }

            const freq = Duration(milliseconds: 50);

            test("${ServiceEndpointVerifier.stateConstruct}", () async {
                await ServiceEndpointVerifier.stateConstruct(
                    RestServiceEndpoint<TestModel>(okModel(), fetchFrequency: freq),
                );
            });

            test("${ServiceEndpointVerifier.stateConstructAndDispose}", () async {
                await ServiceEndpointVerifier.stateConstructAndDispose(
                    RestServiceEndpoint<TestModel>(okModel(), fetchFrequency: freq),
                );
            });

            test("${ServiceEndpointVerifier.stateListenOnce}", () async {
                await ServiceEndpointVerifier.stateListenTwice(
                    RestServiceEndpoint<TestModel>(okModel(), fetchFrequency: freq),
                );
            });

            test("${ServiceEndpointVerifier.stateListenReceivesError}", () async {
                await ServiceEndpointVerifier.stateListenReceivesError(
                    RestServiceEndpoint<TestModel>(() {
                        final model = _MockRAWModelUpdate();
                        when(model.request())
                            .thenAnswer((_) async {
                            return Left(statusError(1234));
                        });
                        return model;
                    }(), fetchFrequency: freq),
                    statusError(1234),
                );
            });

            test("${ServiceEndpointVerifier.stateUnlistenInitializes}", () async {
                await ServiceEndpointVerifier.stateUnlistenInitializes(
                    RestServiceEndpoint<TestModel>(okModel(), fetchFrequency: freq),
                );
            });

            test("${ServiceEndpointVerifier.stateListenTwice}", () async {
                await ServiceEndpointVerifier.stateListenTwice(
                    RestServiceEndpoint<TestModel>(okModel(), fetchFrequency: freq),
                );
            });

            test("${ServiceEndpointVerifier.stateDisposeWhileConnected}", () async {
                await ServiceEndpointVerifier.stateDisposeWhileConnected(
                    RestServiceEndpoint<TestModel>(okModel(), fetchFrequency: freq)
                );
            });

            test("${ServiceEndpointVerifier.stateDisposeImmediatelyAfterConnecting}", () async {
                await ServiceEndpointVerifier.stateDisposeImmediatelyAfterConnecting(
                    RestServiceEndpoint<TestModel>(okModel(), fetchFrequency: freq)
                );
            });

            test("${ServiceEndpointVerifier.stateDisposeAfterUnlistening}", () async {
                await ServiceEndpointVerifier.stateDisposeAfterUnlistening(
                    RestServiceEndpoint<TestModel>(okModel(), fetchFrequency: freq)
                );
            });

            test("${ServiceEndpointVerifier.stateNoInternetOnceThenNoListeners}", () async {
                final responses = [
                    Right<AllErrors, Tuple2<TestModel, String>>(Tuple2(TestModel(), TestModel().convert().to())),
                    const Left<AllErrors, Tuple2<TestModel, String>>(noInternetError),
                ];

                final model = _MockRAWModelUpdate();
                when(model.request()).thenAnswer((_) async => responses.removeLast());

                await ServiceEndpointVerifier.stateNoInternetOnceThenNoListeners(
                    RestServiceEndpoint<TestModel>(model, fetchFrequency: freq)
                );
            });

            test("${ServiceEndpointVerifier.stateReconnectWithNoListeners}", () async {
                final model = _MockRAWModelUpdate();
                when(model.request()).thenAnswer((_) async => throw Exception("Never call"));

                await ServiceEndpointVerifier.stateReconnectWithNoListeners(
                    RestServiceEndpoint<TestModel>(model, fetchFrequency: freq)
                );
            });

            test("${ServiceEndpointVerifier.dataConstruct}", () async {
                final model = _MockRAWModelUpdate();
                when(model.request()).thenAnswer((_) async => throw Exception("Never call"));

                await ServiceEndpointVerifier.dataConstruct(
                    RestServiceEndpoint<TestModel>(model, fetchFrequency: freq)
                );
            });

            test("${ServiceEndpointVerifier.dataSingleListen}", () async {
                await ServiceEndpointVerifier.dataSingleListen(
                    RestServiceEndpoint<TestModel>(okModel(), fetchFrequency: freq),
                    TestModel(),
                );
            });

            test("${ServiceEndpointVerifier.dataSingleListenError}", () async {
                await ServiceEndpointVerifier.dataSingleListenError(
                    RestServiceEndpoint<TestModel>(() {
                        final model = _MockRAWModelUpdate();
                        when(model.request())
                            .thenAnswer((_) async {
                            return Left(statusError(1234));
                        });
                        return model;
                    }(), fetchFrequency: freq),
                    statusError(1234),
                );
            });

            test("${ServiceEndpointVerifier.dataConstructDispose}", () async {
                await ServiceEndpointVerifier.dataConstructDispose(
                    RestServiceEndpoint<TestModel>(okModel(), fetchFrequency: freq),
                );
            });

            test("${ServiceEndpointVerifier.dataListenSuccessDispose}", () async {
                await ServiceEndpointVerifier.dataListenSuccessDispose(
                    RestServiceEndpoint<TestModel>(okModel(), fetchFrequency: freq),
                    TestModel(),
                );
            });

            test("${ServiceEndpointVerifier.dataListenErrorDispose}", () async {
                await ServiceEndpointVerifier.dataListenErrorDispose(
                    RestServiceEndpoint<TestModel>(() {
                        final model = _MockRAWModelUpdate();
                        when(model.request())
                            .thenAnswer((_) async {
                            return Left(statusError(1234));
                        });
                        return model;
                    }(), fetchFrequency: freq),
                    statusError(1234),
                );
            });

            test("${ServiceEndpointVerifier.dataDispose}", () async {
                await ServiceEndpointVerifier.dataDispose(
                    RestServiceEndpoint<TestModel>(okModel(), fetchFrequency: freq),
                );
            });
        });

        group("other", () {
            RestServiceEndpoint<TestModel> sut;
            RAWModelUpdate<TestModel, String> model;
            List<Either<AllErrors, Tuple2<TestModel, String>>> data;

            setUp(() {
                model = _MockRAWModelUpdate();
                data = [
                    Right<AllErrors, Tuple2<TestModel, String>>(Tuple2(TestModel(), TestModel().convert().to())),
                    Right<AllErrors, Tuple2<TestModel, String>>(Tuple2(TestModel(), TestModel().convert().to())),
                    Left<AllErrors, Tuple2<TestModel, String>>(statusError(123)),
                ];

                when(model.request()).thenAnswer((_) async => data.removeAt(0));

                sut = RestServiceEndpoint<TestModel>(
                    model,
                    fetchFrequency: const Duration(milliseconds: 50),
                );
            });

            group("restartObservable", () {
                test("test", () async {
                    expect(sut.lastState.asStream(), emitsInOrder(<dynamic>[
                        apiInitialized,

                        /// b, only change to a loading state on the first time
                        apiLoading,
                        apiConnected,

                        /// c, change state to disconnected when fetched data was an error
                        apiDisconnected(statusError(123)),
                    ]));

                    expect(sut.lastData.asStream(), emitsInOrder(<dynamic>[
                        const None<dynamic>(),
                        Some(Right<dynamic, dynamic>(TestModel())),

                        /// e, restarts observable when there are listeners
                        Some(Right<dynamic, dynamic>(TestModel())),
                        Some(Left<dynamic, dynamic>(statusError(123))),
                    ]));

                    await Future<void>.delayed(const Duration(milliseconds: 300));

                    await sut.dispose();
                });

                test("do nothing when already running", () async {
                    expect(sut.lastState.asStream(), emitsInOrder(<dynamic>[
                        ...apiInitLoadConnect,
                        apiDisconnected(cancelledError),
                        emitsDone,
                    ]));
                    sut.lastData.subscribe();
                    await sut.reconnect();
                    await sut.dispose();
                });

                test("go back to being initialized when there are no listeners and dont restart this routine.", () async {
                    expect(sut.lastState.asStream(), emitsInOrder(<dynamic>[
                        ...apiInitLoadConnect,
                        apiInitialized,
                        apiDisconnected(cancelledError),
                        emitsDone,
                    ]));
                    final sub = sut.lastData.subscribe();
                    await pumpEventQueue();
                    sub.cancel();
                    expect(data.length, 2);
                    await pumpEventQueue();
                    await sut.dispose();
                });
            });

            test("fetchLast", () async {
                when(model.request()).thenAnswer((_) async => Right(Tuple2(TestModel(), TestModel().convert().to())));
                final a = await sut.singleLastData;
                expect(a, Right<AllErrors, TestModel>(TestModel()));
            });

            group("observeLastState", () {
                test("immediately return the last value and only return distinct values", () async {
                    expect(sut.lastState.asStream(), emitsInOrder(<dynamic>[
                        apiInitialized,

                        /// b, only change to a loading state on the first time
                        apiLoading,
                        apiConnected,

                        /// c, change state to disconnected when fetched data was an error
                        apiDisconnected(statusError(123)),
                    ]));

                    expect(sut.lastState.asStream(), emitsInOrder(<dynamic>[
                        ...apiInitLoadConnect,
                        apiDisconnected(statusError(123)),
                        emitsDone,
                    ]));

                    expect(sut.lastData.asStream(), emitsInOrder(<dynamic>[
                        const None<Either>(),
                        Some(Right<dynamic, dynamic>(TestModel())),
                        Some(Right<dynamic, dynamic>(TestModel())),
                        Some(Left<dynamic, dynamic>(statusError(123))),
                    ]));

                    await Future<void>.delayed(const Duration(milliseconds: 200));

                    expect(sut.lastState.asStream(), emitsInOrder(<dynamic>[
                        apiDisconnected(statusError(123)),
                    ]));

                    await sut.dispose();
                });
            });

            group("reconnect", () {
                RestServiceEndpoint<TestModel> _sut;
                RAWModelUpdate<TestModel, String> _model;
                List<Either<AllErrors, Tuple2<TestModel, String>>> _data;

                setUp(() {
                    _model = _MockRAWModelUpdate();
                    _data = [
                        Right<AllErrors, Tuple2<TestModel, String>>(Tuple2(TestModel(), TestModel().convert().to())),
                        Right<AllErrors, Tuple2<TestModel, String>>(Tuple2(TestModel(), TestModel().convert().to())),
                        Left<AllErrors, Tuple2<TestModel, String>>(statusError(123)),
                        Right<AllErrors, Tuple2<TestModel, String>>(Tuple2(TestModel(), TestModel().convert().to())),
                        Left<AllErrors, Tuple2<TestModel, String>>(statusError(123)),
                    ];

                    when(_model.request()).thenAnswer((_) async => _data.removeAt(0));

                    _sut = RestServiceEndpoint<TestModel>(
                        _model,
                        fetchFrequency: const Duration(milliseconds: 50),
                    );
                });

                test("reconnect with one listener after error thrown", () async {
                    expect(_sut.lastState.asStream(), emitsInOrder(<dynamic>[
                        ...apiInitLoadConnect,
                        apiDisconnected(statusError(123)),
                        ...apiInitLoadConnect,
                        apiDisconnected(statusError(123)),
                    ]));

                    expect(_sut.lastData.asStream(), emitsInOrder(<dynamic>[
                        const None<Either>(),
                        Some(Right<dynamic, dynamic>(TestModel())),
                        Some(Right<dynamic, dynamic>(TestModel())),
                        Some(Left<dynamic, dynamic>(statusError(123))),
                        Some(Right<dynamic, dynamic>(TestModel())),
                        Some(Left<dynamic, dynamic>(statusError(123))),
                    ]));

                    _sut.lastData.subscribe((a) {});

                    await Future<void>.delayed(const Duration(milliseconds: 300));

                    await _sut.reconnect();

                    await Future<void>.delayed(const Duration(milliseconds: 300));
                });
            });
        });
    });

    group("$StaticServiceEndpointImpl", () {
        StaticServiceEndpointImpl sut;

        _MockServiceEndpoint rawEndpoint;

        // ignore: close_sinks
        Signal<Option<Either<AllErrors, String>>> data;

        // ignore: close_sinks
        Signal<APIState> state;
        setUp(() {
            data = Signal(const None());
            state = Signal(apiInitialized);
            rawEndpoint = _MockServiceEndpoint();
            when(rawEndpoint.lastData).thenAnswer((_) => data.wave.as(OEWave.from));
            when(rawEndpoint.lastState).thenAnswer((_) => state.wave);
            sut = StaticServiceEndpointImpl<dynamic>(rawEndpoint, true);
        });

        tearDown(() {
            verifyNever(rawEndpoint.dispose());
        });

        group("state", () {
            test("seeded to initialized", () async {
                expect(sut.lastState.value, apiInitialized);
            });
        });

        group("update & data", () {
            test("error", () async {
                expect(sut.lastState.asStream(), emitsInOrder(<dynamic>[
                    apiInitialized,
                    apiLoading,
                    apiDisconnected(unknownAPIError),
                ]));

                expect(sut.lastData.asStream(onListen: () {}), emitsInOrder(<dynamic>[
                    const None<Either>(),
                    const Some(Left<dynamic, dynamic>(unknownAPIError)),
                ]));

                data.add(const Some(Left(unknownAPIError)));

                verify(rawEndpoint.lastData).called(2);
            });

            test("data received", () {
                expect(sut.lastState.asStream(), emitsInOrder(<dynamic>[
                    ...apiInitLoadConnect
                ]));

                expect(sut.lastData.asStream(), emitsInOrder(<dynamic>[
                    const None<Either>(),
                    const Some(Right<dynamic, dynamic>("abc")),
                ]));

                data.add(const Some(Right("abc")));

                verify(rawEndpoint.lastData).called(2);
            });

            test("don't update when loading", () async {
                expect(sut.lastState.asStream(), emitsInOrder(<dynamic>[
                    ...apiInitLoadConnect,
                ]));

                expect(sut.lastData.asStream(), emitsInOrder(<dynamic>[
                    const None<Either>(),
                    const Some(Right<dynamic, dynamic>("abc")),
                ]));

                await sut.update();
                await sut.update();
                await sut.update();

                data.add(const Some(Right("abc")));

                verify(rawEndpoint.lastData).called(2);
            });

            test("do nothing when no listeners", () async {
                expect(sut.lastState.asStream(), emitsInOrder(<dynamic>[
                    apiInitialized,
                ]));

                await sut.update();
                verify(rawEndpoint.lastData).called(1);
            });
        });
    });
}

class _MockRAWModelUpdate extends Mock implements RAWModelUpdate<TestModel, String> {}

class TestModel extends SmartJSON<TestModel> {

    final SmartString a = SmartString("aa");

    final SmartString b = SmartString("bb");

    @override
    TestModel creator() => TestModel();

    @override
    JSONPrototypeCodec<TestModel> tactic() => JSONMapTactic.ignoreMissing({"a": a, "b": b});
}

class _MockServiceEndpoint extends Mock implements ServiceEndpoint<String> {}