import 'dart:async';

import 'package:bird/smart_api.dart';
import 'package:dartz/dartz.dart';
import 'package:test/test.dart';

/// This file holds verification routines for [ServiceEndpoint] implementations.
///
/// All implementations have to follow this spec.
///
/// An implementation that successfully passes all these tests
/// is not necessarily correct. More tests are needed that verify
/// each implementations concrete behavior.
class ServiceEndpointVerifier {

    // State tests

    /// The state starts out initialized.
    ///
    /// Tests:
    /// Endpoint starts out with [APIStateInitialized] for the state
    /// and [None] for the data
    static Future<void> stateConstruct<T>(ServiceEndpointStreams<T> sut) async {
        expect(sut.lastState.asStream(), emitsInOrder(<dynamic>[
            apiInitialized,
        ]));
    }

    /// The state starts out initialized and is then disposed.
    ///
    /// Tests:
    /// Disposing leads to [APIStateDisconnected] with [CancelledAPIError].
    static Future<void> stateConstructAndDispose<T>(ServiceEndpointStreams<T> sut) async {
        expect(sut.lastState.asStream(), emitsInOrder(<dynamic>[
            apiInitialized,
            apiDisconnected(cancelledError),
            emitsDone,
        ]));
        await sut.dispose();
    }

    /// Listening for data once (successfully) goes through the usual state changes.
    ///
    /// Tests:
    /// Listening leads to
    /// [APIStateInitialized] -> [APIStateLoading] -> [APIStateConnected].
    static Future<void> stateListenOnce<T>(ServiceEndpointStreams<T> sut) async {
        expect(sut.lastState.asStream(), emitsInOrder(<dynamic>[
            ...apiInitLoadConnect,
        ]));
        sut.lastData.subscribe();
    }

    /// Listening for data once (successfully) goes through the usual state changes.
    ///
    /// Tests:
    /// Listening leads to
    /// [APIStateInitialized] -> [APIStateLoading] -> [APIStateDisconnected].
    static Future<void> stateListenReceivesError<T>(ServiceEndpointStreams<T> sut, AllErrors error) async {
        expect(sut.lastState.asStream(), emitsInOrder(<dynamic>[
            apiInitialized,
            apiLoading,
            mayEmit(apiConnected),
            apiDisconnected(error),
        ]));
        sut.lastData.subscribe();
    }

    /// After [firstSuccessfulListenState], if all listeners stop listening
    /// for updates, the state goes back to being [APIStateInitialized].
    ///
    /// Tests:
    /// Unsub while connected leads to [APIStateInitialized].
    static Future<void> stateUnlistenInitializes<T>(ServiceEndpointStreams<T> sut) async {
        expect(sut.lastState.asStream(), emitsInOrder(<dynamic>[
            ...apiInitLoadConnect,
            apiInitialized,
        ]));
        final sub = sut.lastData.subscribe((data) {});
        await pumpEventQueue();
        sub.cancel();
        await pumpEventQueue();
    }

    /// Listening for data twice (successfully) goes through the usual state changes.
    /// The first listener cancelling its subscription should not have no influence.
    /// Tests:
    ///
    /// Listening leads to
    /// [APIStateInitialized] -> [APIStateLoading] -> [APIStateConnected].
    static Future<void> stateListenTwice<T>(ServiceEndpointStreams<T> sut) async {
        expect(sut.lastState.asStream(), emitsInOrder(<dynamic>[
            ...apiInitLoadConnect,
        ]));
        final a = sut.lastData.subscribe();
        sut.lastData.subscribe();
        a.cancel();
    }

    /// If the service endpoint is disposed while being connected it
    /// will switch to being [APIStateDisconnected] with a [CancelledAPIError].
    ///
    /// Tests:
    /// Disposal while connected leads to [APIStateDisconnected] with a [CancelledAPIError].
    static Future<void> stateDisposeWhileConnected<T>(ServiceEndpointStreams<T> sut) async {
        expect(sut.lastState.asStream(), emitsInOrder(<dynamic>[
            ...apiInitLoadConnect,
            apiDisconnected(cancelledError),
            emitsDone,
        ]));
        sut.lastData.subscribe();
        await pumpEventQueue();
        await sut.dispose();
    }

    /// immediately disposing after starting to connect skips the connected state because
    /// there is not enough time to go to Future execution
    ///
    ///
    /// Tests:
    /// Disposal after listening to data without pumping the event queue skips the
    /// [APIStateConnected] state.
    //  this may be too restrictive for other implementations
    static Future<void> stateDisposeImmediatelyAfterConnecting<T>(ServiceEndpointStreams<T> sut) async {
        expect(sut.lastState.asStream(), emitsInOrder(<dynamic>[
            apiInitialized,
            apiLoading,
            mayEmit(apiConnected),
            apiDisconnected(cancelledError),
            emitsDone,
        ]));
        sut.lastData.subscribe();
        await sut.dispose();
    }

    /// A cancelled subscription leads to an [APIStateInitialized] state and the
    /// immediate disposal of the endpoint thereafter to [APIStateDisconnected]
    /// with a [CancelledAPIError].
    ///
    /// Tests:
    /// Disposal after a cancelled subscription leads to [APIStateDisconnected]
    /// with a [CancelledAPIError].
    static Future<void> stateDisposeAfterUnlistening<T>(ServiceEndpointStreams<T> sut) async {
        expect(sut.lastState.asStream(), emitsInOrder(<dynamic>[
            ...apiInitLoadConnect,
            apiInitialized,
            apiDisconnected(cancelledError),
            emitsDone,
        ]));
        final sub = sut.lastData.subscribe();
        await pumpEventQueue();
        sub.cancel();
        await sut.dispose();
    }

    /// Tries to get data once but there is no internet connection.
    /// Reconnects again receives data successfully and stops
    /// listening (while receiving new data) which leads the
    /// state to being initialized.
    ///
    /// Tests:
    /// Unsub right after reconnecting.
    static Future<void> stateNoInternetOnceThenNoListeners<T>(ServiceEndpoint<T> sut) async {
        expect(sut.lastState.asStream(), emitsInOrder(<dynamic>[
            apiInitialized,
            apiLoading,
            mayEmit(apiConnected),
            apiDisconnected(noInternetError),
            ...apiInitLoadConnect,
            apiInitialized,
        ]));

        final subscription = sut.lastData.subscribe();
        await pumpEventQueue();
        sut.reconnect();
        await pumpEventQueue();
        subscription.cancel();
    }

    /// Calling reconnect with no listeners should do nothing
    ///
    /// Tests:
    /// Reconnect does nothing if there are no listeners.
    static Future<void> stateReconnectWithNoListeners<T>(ServiceEndpoint<T> sut) async {
        expect(sut.lastState.asStream(), emitsInOrder(<dynamic>[
            apiInitialized,
            apiDisconnected(cancelledError),
            emitsDone,
        ]));
        await pumpEventQueue();
        sut.reconnect();
        await pumpEventQueue();
        await sut.dispose();
        await pumpEventQueue();
    }

    // Data tests

    /// Data starts out empty
    ///
    /// Tests:
    /// Data is None when constructed
    static Future<void> dataConstruct<T>(ServiceEndpointStreams<T> sut) async {
        expect(sut.lastData.value, None<Either<AllErrors, T>>());
    }

    /// Data starts out empty and returns some data when listened to
    ///
    /// Tests:
    /// First listen gives data.
    static Future<void> dataSingleListen<T>(ServiceEndpointStreams<T> sut, T t) async {
        expect(sut.lastData.asStream(), emitsInOrder(<dynamic>[
            None<Either<AllErrors, T>>(),
            Some(Right<AllErrors, T>(t)),
        ]));
    }

    /// Data starts out empty and returns an error.
    ///
    /// Tests:
    /// First listen gives error.
    static Future<void> dataSingleListenError<T>(ServiceEndpointStreams<T> sut, AllErrors error) async {
        expect(sut.lastData.asStream(), emitsInOrder(<dynamic>[
            None<Either<AllErrors, T>>(),
            Some(Left<AllErrors, T>(error)),
        ]));
    }

    /// Data starts out empty and returns an error.
    ///
    /// Tests:
    /// First listen gives error.
    static Future<void> dataConstructDispose<T>(ServiceEndpointStreams<T> sut) async {
        expect(sut.lastData.asStream(), emitsInOrder(<dynamic>[
            None<Either<AllErrors, T>>(),
            emitsDone,
        ]));
        await sut.dispose();
    }

    /// Successfully receives data once and disposes right after
    ///
    /// Tests:
    /// No data is emitted when disposed.
    static Future<void> dataListenSuccessDispose<T>(ServiceEndpointStreams<T> sut, T data) async {
        expect(sut.lastData.asStream(), emitsInOrder(<dynamic>[
            None<Either<AllErrors, T>>(),
            Some(Right<AllErrors, T>(data)),
            emitsDone,
        ]));
        await pumpEventQueue();
        await sut.dispose();
    }

    /// Successfully receives data once and disposes right after
    ///
    /// Tests:
    /// No data is emitted when disposed.
    static Future<void> dataListenErrorDispose<T>(ServiceEndpointStreams<T> sut, AllErrors error) async {
        expect(sut.lastData.asStream(), emitsInOrder(<dynamic>[
            None<Either<AllErrors, T>>(),
            Some(Left<AllErrors, T>(error)),
            emitsDone,
        ]));
        await pumpEventQueue();
        await sut.dispose();
    }

    /// Disposes endpoint immediately without listening for data.
    ///
    /// Tests:
    /// Data starts out as none and stays none after disposal.
    static Future<void> dataDispose<T>(ServiceEndpointStreams<T> sut) async {
        expect(sut.lastData.value, None<Either<AllErrors, T>>());
        await pumpEventQueue();
        await sut.dispose();
        expect(sut.lastData.value, None<Either<AllErrors, T>>());
    }

    // reconnect

}