import 'dart:convert';

import 'package:bird/smart_api.dart';
import 'package:collection/collection.dart';
import 'package:dartz/dartz.dart';
import 'package:test/test.dart';

void expectNone<T>(Option<T> value) {
    expect(value, None<T>());
}

void expectSome<T>(Option<T> value, T t) {
    expect(value, Some(t));
}

void main() {
    group("$JSONRPC2_0RequestObject", () {
        group("asserts", () {
            test("method not null", () {
                expect(
                        () => JSONRPC2_0RequestObject(method: null, id: 5, params: [123, "123", true]),
                    throwsA(const TypeMatcher<AssertionError>()));
            });
            test("params list or map or null", () {
                expect(
                        () => JSONRPC2_0RequestObject(method: "null", id: 5, params: "invalid"),
                    throwsA(const TypeMatcher<AssertionError>()));
            });
        });

        test(".notification", () {
            final sut = JSONRPC2_0RequestObject.notification(method: "method", params: {"a": "b"});
            expect(sut.method, "method");
            expect(sut.params.getOrElse(() => null), {"a": "b"});
        });

        test("fromJson", () {
            expect(JSONRPC2_0RequestObject.fromJson(<String, dynamic>{"method": "subtract", "params": [42, 23], "id": 1}),
                isA<Left<RPCException, JSONRPC2_0RequestObject>>());

            expect(
                JSONRPC2_0RequestObject.fromJson(
                    <String, dynamic>{"jsonrpc": "2.1", "method": "method", "params": [42, 23], "id": 1}),
                isA<Left<RPCException, JSONRPC2_0RequestObject>>());

            expect(JSONRPC2_0RequestObject.fromJson(<String, dynamic>{"jsonrpc": "2.0", "params": [42, 23], "id": 1}),
                isA<Left<RPCException, JSONRPC2_0RequestObject>>());

            expect(
                JSONRPC2_0RequestObject.fromJson(
                    <String, dynamic>{"jsonrpc": "2.0", "method": "subtract", "params": "a", "id": 1}),
                isA<Left<RPCException, JSONRPC2_0RequestObject>>());

            final sut = JSONRPC2_0RequestObject.fromJson(
                <String, dynamic>{"jsonrpc": "2.0", "method": "subtract", "params": [42, 23], "id": 1}) | null;
            expect(sut.method, "subtract");
            expect(sut.params.getOrElse(() => null), [42, 23]);
            expect(sut.id.getOrElse(() => null), 1);

            final a = JSONRPC2_0RequestObject.fromJson(<String, dynamic>{
                JSON_RPC_Constants.JSONRPC_KEY: JSON_RPC_Constants.JSON_RPC_2_0_VERSION_VALUE,
                JSON_RPC_Constants.METHOD_KEY: "null",
                JSON_RPC_Constants.ID_KEY: 5,
                JSON_RPC_Constants.PARAMS_KEY: [123, "123", true],
            }) | null;

            expect(a.method, "null");
            expect(a.params.getOrElse(() => null), [123, "123", true]);
            expect(a.id, some(5));
        });

        test("toJson", () {
            expect(
                JSONRPC2_0RequestObject(method: "null", id: 5, params: [123, "123", true]).toJson(),
                {
                    JSON_RPC_Constants.JSONRPC_KEY: JSON_RPC_Constants.JSON_RPC_2_0_VERSION_VALUE,
                    JSON_RPC_Constants.METHOD_KEY: "null",
                    JSON_RPC_Constants.ID_KEY: 5,
                    JSON_RPC_Constants.PARAMS_KEY: [123, "123", true],
                });
        });

        test("toString", () {
            expect(
                json.decode(JSONRPC2_0RequestObject(method: "null", id: 5, params: [123, "123", true]).toString()),
                {
                    JSON_RPC_Constants.JSONRPC_KEY: JSON_RPC_Constants.JSON_RPC_2_0_VERSION_VALUE,
                    JSON_RPC_Constants.METHOD_KEY: "null",
                    JSON_RPC_Constants.ID_KEY: 5,
                    JSON_RPC_Constants.PARAMS_KEY: [123, "123", true],
                });
        });

        group("some JSONRPC2.0 request object specification tests", () {
            test("with positional parameters", () {
                final a = JSONRPC2_0RequestObject.fromJson(
                    <String, dynamic>{"jsonrpc": "2.0", "method": "subtract", "params": [42, 23], "id": 1}) | null;
                expect(a.method, "subtract");
                expect(a.params.getOrElse(() => null), [42, 23]);
                expect(a.id, some(1));
            });

            test("with named parameters", () {
                final a = JSONRPC2_0RequestObject.fromJson(
                    <String, dynamic>{
                        "jsonrpc": "2.0",
                        "method": "subtract",
                        "params": {"subtrahend": 23, "minuend": 42},
                        "id": 3
                    }) | null;
                expect(a.method, "subtract");
                expect(a.params.getOrElse(() => null), {"subtrahend": 23, "minuend": 42});
                expect(a.id, some(3));
            });

            test("with named parameters", () {
                final a = JSONRPC2_0RequestObject.fromJson(
                    <String, dynamic>{
                        "jsonrpc": "2.0",
                        "method": "subtract",
                        "params": {"subtrahend": 23, "minuend": 42},
                        "id": 3
                    }) | null;
                expect(a.method, "subtract");
                expect(a.params.getOrElse(() => null), {"subtrahend": 23, "minuend": 42});
                expect(a.id, some(3));
            });

            test("a notification", () {
                final a = JSONRPC2_0RequestObject.fromJson(
                    <String, dynamic>{"jsonrpc": "2.0", "method": "update", "params": [1, 2, 3, 4, 5]}) | null;
                expect(a.method, "update");
                expect(a.params.getOrElse(() => null), [1, 2, 3, 4, 5]);
                expect(a.id, none<Object>());
            });
        });
    });

    test("$MissingRPCVersionException", () {
        expect(const MissingRPCVersionException("-").toString(), contains("$MissingRPCVersionException"));
    });

    test("$WrongRPCVersionException", () {
        expect(const WrongRPCVersionException("-").toString(), contains("$WrongRPCVersionException"));
    });

    test("$MissingMethodRPCException", () {
        expect(const MissingMethodRPCException("-").toString(), contains("$MissingMethodRPCException"));
    });

    test("$BadParamsRPCException", () {
        expect(const BadParamsRPCException("-").toString(), contains("$BadParamsRPCException"));
    });

    group("$JSONRPC2_0ResponseObject", () {
        test("only error or result at a time", () {
            expect(
                    () {
                    return JSONRPC2_0ResponseObject(
                        result: none<dynamic>(),
                        error: none(),
                        id: 3,
                    );
                },
                throwsA(const TypeMatcher<AssertionError>()),
            );
            expect(
                    () {
                    return JSONRPC2_0ResponseObject(
                        result: some<dynamic>(<String, dynamic>{}),
                        error: some(JSON_RPC_2_Error.internalError()),
                        id: 3,
                    );
                },
                throwsA(const TypeMatcher<AssertionError>()),
            );
        });
        test("fromJson", () {
            expect(JSONRPC2_0ResponseObject.fromJson(<String, dynamic>{"result": 19, "id": 1}),
                isA<Either<RPCException, JSONRPC2_0ResponseObject>>());
            expect(JSONRPC2_0ResponseObject.fromJson(<String, dynamic>{"jsonrpc": "2.1", "result": 19, "id": 1}),
                isA<Left<RPCException, JSONRPC2_0ResponseObject>>());
            expect(JSONRPC2_0ResponseObject.fromJson(<String, dynamic>{"jsonrpc": "2.0", "id": 1}),
                isA<Left<RPCException, JSONRPC2_0ResponseObject>>());
            expect(JSONRPC2_0ResponseObject.fromJson(<String, dynamic>{"jsonrpc": "2.0", "result": 19}),
                isA<Left<RPCException, JSONRPC2_0ResponseObject>>());

            final sut = JSONRPC2_0ResponseObject.fromJson(<String, dynamic>{"jsonrpc": "2.0", "result": 19, "id": 1}) | null;

            expect(sut.result, some(19));
            expect(sut.id, 1);
            expect(sut.error, const None<JSON_RPC_2_Error>());
        });

        test("toJson", () {
            final sut = JSONRPC2_0ResponseObject.fromJson(<String, dynamic>{"jsonrpc": "2.0", "result": 19, "id": 1}) | null;
            expect(sut.toJson(), {"jsonrpc": "2.0", "result": 19, "id": 1});
        });

        test("toJson", () {
            final sut = JSONRPC2_0ResponseObject.fromJson(
                <String, dynamic>{"jsonrpc": "2.0", "error": JSON_RPC_2_Error.internalError().toJson(), "id": 1}) | null;
            expect(
                sut.toJson(), <String, dynamic>{"jsonrpc": "2.0", "error": JSON_RPC_2_Error.internalError().toJson(), "id": 1});
        });

        group("some JSONRPC2.0 response object specification tests", () {
            test("normal usage", () {
                final sut = JSONRPC2_0ResponseObject.fromJson(<String, dynamic>{"jsonrpc": "2.0", "result": 19, "id": 1}) | null;
                expect(sut.result, some(19));
                expect(sut.id, 1);
                expectNone(sut.error);
            });

            test("parseError", () {
                final sut = JSONRPC2_0ResponseObject.fromJson(
                    <String, dynamic>{"jsonrpc": "2.0", "error": {"code": -32700, "message": "Parse error"}, "id": null}) | null;
                expectNone<dynamic>(sut.result);
                expect(sut.id, null);
                expect(sut.error, some(JSON_RPC_2_Error.parseError()));
            });

            test("methodNotFound", () {
                final sut = JSONRPC2_0ResponseObject.fromJson(
                    <String, dynamic>{
                        "jsonrpc": "2.0",
                        "error": {"code": -32601, "message": "Method not found"},
                        "id": "1"
                    }) | null;
                expectNone<dynamic>(sut.result);
                expect(sut.id, "1");
                expect(sut.error, some(JSON_RPC_2_Error.methodNotFound()));
            });

            test("invalidRequest", () {
                final sut = JSONRPC2_0ResponseObject.fromJson(
                    <String, dynamic>{
                        "jsonrpc": "2.0",
                        "error": {"code": -32600, "message": "Invalid Request"},
                        "id": null
                    }) | null;
                expectNone<dynamic>(sut.result);
                expect(sut.id, null);
                expect(sut.error, some(JSON_RPC_2_Error.invalidRequest()));
            });
        });

        test("toString", () {
            final Either<RPCException, JSONRPC2_0ResponseObject> sut = JSONRPC2_0ResponseObject.fromJson(<String, dynamic>{"jsonrpc": "2.0", "result": 19, "id": 1});
            expect(json.decode((sut | null).toString()), {"jsonrpc": "2.0", "result": 19, "id": 1});
        });
    });

    test("$NoContentRPCException", () {
        expect(const NoContentRPCException("-").toString(), contains("$NoContentRPCException"));
    });

    test("$NoIDRPCException", () {
        expect(const NoIDRPCException("-").toString(), contains("$NoIDRPCException"));
    });

    group("$JSON_RPC_2_Error", () {
        test("parse error", () {
            final sut = JSON_RPC_2_Error.parseError();
            expect(sut.code, -32700);
            expect(sut.message, "Parse error");
        });

        test("invalid request", () {
            final sut = JSON_RPC_2_Error.invalidRequest();
            expect(sut.code, -32600);
            expect(sut.message, "Invalid Request");
        });

        test("method not found", () {
            final sut = JSON_RPC_2_Error.methodNotFound();
            expect(sut.code, -32601);
            expect(sut.message, "Method not found");
        });

        test("invalid params", () {
            final sut = JSON_RPC_2_Error.invalidParams();
            expect(sut.code, -32602);
            expect(sut.message, "Invalid params");
        });

        test("internal error", () {
            final sut = JSON_RPC_2_Error.internalError();
            expect(sut.code, -32603);
            expect(sut.message, "Internal error");
        });

        test("server error", () {
            final sut = JSON_RPC_2_Error.serverError(0);
            expect(sut.code, -32000);
            expect(sut.message, "Server error");

            final sut2 = JSON_RPC_2_Error.serverError(99);
            expect(sut2.code, -32099);
            expect(sut2.message, "Server error");

            expect(() => JSON_RPC_2_Error.serverError(100), throwsA(const TypeMatcher<AssertionError>()));
            expect(() => JSON_RPC_2_Error.serverError(-1), throwsA(const TypeMatcher<AssertionError>()));
        });

        test("server error", () {
            final sut = JSON_RPC_2_Error.application(-30000, "a");
            expect(sut.code, -30000);
            expect(sut.message, "a");
        });

        test("application", () {
            final a = JSON_RPC_2_Error.application(-1, "a", data: const None<dynamic>());
            expect(a.code, -1);
            expect(a.message, "a");
            expectNone<dynamic>(a.data);
        });

        group("fromJson", () {
            test("doesn't contain code", () {
                expect(JSON_RPC_2_Error.fromJSON(<String, dynamic>{}), JSON_RPC_2_Error.parseError());
            });

            test("doesn't contain message", () {
                expect(JSON_RPC_2_Error.fromJSON({"code": -123}), JSON_RPC_2_Error.parseError());
            });

            test("not map", () {
                expect(JSON_RPC_2_Error.fromJSON(true), JSON_RPC_2_Error.parseError());
            });

            test("valid", () {
                expect(JSON_RPC_2_Error.fromJSON({"code": -123, "message": "msg"}),
                    const JSON_RPC_2_Error(-123, "msg", None<dynamic>()));
            });
        });

        group("toJson", () {
            test("no data", () {
                final sut = JSON_RPC_2_Error.internalError();
                expect(json.encode(sut.toJson()), '{"code":-32603,"message":"Internal error"}');
            });

            test("with data", () {
                final sut = JSON_RPC_2_Error.internalError(data: some<dynamic>([1]));
                expect(json.encode(sut.toJson()), '{"code":-32603,"message":"Internal error","data":[1]}');
            });
        });

        test("toString", () {
            final sut = JSON_RPC_2_Error.internalError().toString();
            expect(sut.toString(), '{"code":-32603,"message":"Internal error"}');
        });

        test("==", () {
            expect(JSON_RPC_2_Error.internalError(), JSON_RPC_2_Error.internalError());
        });

        test("!=", () {
            expect(JSON_RPC_2_Error.internalError(), isNot(JSON_RPC_2_Error.invalidParams()));
        });

        test("hashCode", () {
            expect(JSON_RPC_2_Error
                .internalError()
                .hashCode, JSON_RPC_2_Error
                .internalError()
                .hashCode);
        });
    });

    group("$mapAnyJsonRpc20Object", () {
        test("A, failed parsing json", () {
            expect(mapAnyJsonRpc20Object<bool>(
                response: (dynamic _, dynamic __) => false,
                request: (_, __, ___) => false,
                error: (_, dynamic __) => false,
                failed: (dynamic _, __) => true,
                data: "{",
            ), true);
        });

        test("B, valid request object", () {
            expect(mapAnyJsonRpc20Object<bool>(
                response: (dynamic _, dynamic __) => false,
                request: (method, params, id) {
                    expect(method, "subtract");
                    expect(params.getOrElse(() => null), [42, 23]);
                    expect(id, some(1));
                    return true;
                },
                error: (dynamic _, dynamic __) => false,
                failed: (dynamic _, dynamic __) => false,
                data: '{"jsonrpc": "2.0", "method": "subtract", "params": [42, 23], "id": 1}',
            ), true);
        });

        test("C, response object without result or error", () {
            expect(mapAnyJsonRpc20Object<bool>(
                response: (dynamic _, dynamic __) => false,
                request: (_, __, ___) => false,
                error: (_, dynamic __) => false,
                failed: (dynamic e, __) {
                    expect(e, const TypeMatcher<Exception>());
                    return true;
                },
                data: '{"jsonrpc": "2.0", "id": 1}',
            ), true);
        });

        test("D, response object with error", () {
            expect(mapAnyJsonRpc20Object<bool>(
                response: (dynamic _, dynamic __) => false,
                request: (_, __, ___) => false,
                error: (error, dynamic id) {
                    expect(id, null);
                    expect(error, JSON_RPC_2_Error.invalidRequest());
                    return true;
                },
                failed: (dynamic _, __) => false,
                data: '{"jsonrpc": "2.0", "error": {"code": -32600, "message": "Invalid Request"}, "id": null}',
            ), true);
        });

        test("E, response object without error", () {
            expect(mapAnyJsonRpc20Object<bool>(
                response: (dynamic result, dynamic id) {
                    expect(result, 19);
                    expect(id, 1);
                    return true;
                },
                request: (_, __, ___) => false,
                error: (_, dynamic __) => false,
                failed: (dynamic _, __) => false,
                data: '{"jsonrpc": "2.0", "result": 19, "id": 1}',
            ), true);
        });

        test("F, neither a response object nor a request object", () {
            expect(mapAnyJsonRpc20Object<bool>(
                response: (dynamic _, dynamic __) => false,
                request: (_, __, ___) => false,
                error: (_, dynamic __) => false,
                failed: (dynamic _, __) => true,
                data: "{}",
            ), true);
        });
    });

    group("$whitelistRpcObject", () {
        test("A", () {
            expect(whitelistRpcObject(
                request: (method, params, id) {
                    expect(method, "subtract");
                    expect(params.getOrElse(() => null), [42, 23]);
                    expect(id, some(1));
                    return true;
                },
                data: '{"jsonrpc": "2.0", "method": "subtract", "params": [42, 23], "id": 1}',
            ), true);
            expect(whitelistRpcObject(
                data: '{"jsonrpc": "2.0", "method": "subtract", "params": [42, 23], "id": 1}',
            ), false);
        });

        test("B", () {
            expect(whitelistRpcObject(
                response: (dynamic result, dynamic id) {
                    expect(result, 19);
                    expect(id, 1);
                    return true;
                },
                data: '{"jsonrpc": "2.0", "result": 19, "id": 1}',
            ), true);
            expect(whitelistRpcObject(
                data: '{"jsonrpc": "2.0", "result": 19, "id": 1}',
            ), false);
        });

        test("C", () {
            expect(whitelistRpcObject(
                error: (error, dynamic id) {
                    expect(id, null);
                    expect(error, JSON_RPC_2_Error.invalidRequest());
                    return true;
                },
                data: '{"jsonrpc": "2.0", "error": {"code": -32600, "message": "Invalid Request"}, "id": null}',
            ), true);
            expect(whitelistRpcObject(
                data: '{"jsonrpc": "2.0", "error": {"code": -32600, "message": "Invalid Request"}, "id": null}',
            ), false);
        });
        test("D", () {
            expect(whitelistRpcObject(
                failed: (dynamic _, __) => true,
                data: "{}",
            ), true);
            expect(whitelistRpcObject(
                data: "{}",
            ), false);
        });
    });

    group("$whitelistRequestRpcObject", () {
        test("A, a", () {
            expect(whitelistRequestRpcObject(
                method: "subtract",
                allowParams: (d) => d is List && const ListEquality<Object>().equals(d, [42, 23]),
                data: '{"jsonrpc": "2.0", "method": "subtract", "params": [42, 23], "id": 1}',
            ), true);
        });

        test("A, b", () {
            expect(whitelistRequestRpcObject(
                method: "subtract",
                allowParams: (d) => d is List && const ListEquality<Object>().equals(d, [42, 23]),
                data: '{"jsonrpc": "2.0", "method": "subtract2", "params": [42, 23], "id": 1}',
            ), false);
        });

        test("A, c", () {
            expect(whitelistRequestRpcObject(
                method: "subtract",
                allowParams: (d) => d is List && const ListEquality<Object>().equals(d, [42, 23]),
                data: '{"jsonrpc": "2.0", "method": "subtract", "params": [42, 24], "id": 1}',
            ), false);
        });

        test("B", () {
            expect(whitelistRequestRpcObject(
                method: "subtract",
                allowParams: (d) => d is List && const ListEquality<Object>().equals(d, [42, 23]),
                data: '{"jsonrpc": "2.0", "result": 19, "id": 1}',
            ), false);
        });

        test("C", () {
            expect(whitelistRequestRpcObject(
                method: "subtract",
                allowParams: (d) => d is List && const ListEquality<Object>().equals(d, [42, 23]),
                data: '{"jsonrpc": "2.0", "error": {"code": -32600, "message": "Invalid Request"}, "id": null}',
            ), false);
        });
        test("D", () {
            expect(whitelistRequestRpcObject(
                method: "subtract",
                allowParams: (d) => d is List && const ListEquality<Object>().equals(d, [42, 23]),
                data: "{}",
            ), false);
        });
    });
}
