import 'dart:convert';

import 'package:bird/smart_json.dart';
import 'package:test/test.dart';

void main() {
  test("SmartModelDict test", () {
    final model = TestModelDict();

    expect(
        model.to(), {"model": <String, dynamic>{"bla": TestModelMap().to()}});
  });

  test("SmartJSONList test", () {
    final model = _TestModelList();

    expect(model.to(), testModelList);

    final model2 = model.copyWith((m) =>
    m
      ..boolVar.safeSet(false)
      ..doubleVar.safeSet(34.0)
      ..intVar.applyIncrement()
      ..stringVar.safeSet("ok2")
    );

    expect(model2.to(), _TestModelList().from([
      false,
      34.0,
      -33,
      "ok2",
    ]).to());
  });

  test("ListOfLists", () {
    final m = _ListOfLists();
    expect(m.to(), [list, list]);
  });

  test("SmartJSON test", () {
    final model = TestModelMap();

    expect(model.to(), testModelMap);

    final model2 = model.copyWith((m) =>
    m
      ..boolVar.safeSet(false)
      ..doubleVar.safeSet(34.0)
      ..intVar.applyIncrement()
      ..stringVar.safeSet("ok2")
      ..boolDictVar.safeApplySet((a) => a..["bla"] = false)
      ..decimal.safeSet("13")
      ..stringDictVar.safeApplySet((a) =>
      a
        ..putIfAbsent("damn", () => "ok")
        ..remove("dictkey")
        ..putIfAbsent("damn2", () => "okok")
        ..remove("damn")),
    );

    expect(model2.to(), TestModelMap().from({
      "bool": false,
      "double": 34.0,
      "int": -33,
      "string": "ok2",
      "stringdict": {
        "damn2": "okok",
      },
      "booldict": {
        "bla": false,
      },
      "emptybooldict": <String, dynamic>{
      },
      "smartdecimal": "13",
    }).to());
  });
  group("smartjson.dart", () {
    group("SmartJSON", () {
      final m = _MockSmartJSON();

      test("creator", () {
        expect(m
            .creator()
            .runtimeType, m.runtimeType);
      });

      test("tactic toJson", () {
        expect(m.tactic().codec(m.creator()).encode(m), {"a": "b"});
        expect(m.tactic().codec(m.creator()).encode(m), isNot({"a": "c"}));
      });

      test("tactic fromJson", () {
        final n = m.tactic().codec(m.creator()).decode({"a": "c"});
        expect(n.str.nullVal(), "c");
        expect(n.to(), {"a": "c"});
      });

      test("fromStringJson valid", () {
        final n = m.stringCodec().decode('{"a": "abc"}');
        expect(n.str.nullVal(), "abc");
        expect(n.to(), {"a": "abc"});
      });

      test("fromStringJson not all values present", () {
        final n = m.stringCodec().decode('{}');
        expect(n.str.nullVal(), null);
        expect(n.to(), {"a": null});
      });

      test("convert().to", () {
        expect(_MockSmartJSON().convert().to(), json.encode({"a": "b"}));
      });
    });

    group("SmartVariableInterface", () {
      /// Tested implicitly
    });

    group("firstOfList", () {
      test("conversion", () {
        expect(
            firstOfList(_TestEnum.values, _TestEnum.a)(_TestEnum.b.toString()),
            _TestEnum.b);
      });
      test("fallback error", () {
        expect(firstOfList(_TestEnum.values, _TestEnum.a)("x"), _TestEnum.a);
      });
    });
  });

  test("fromStringJson with faulty string", () {
    final model = TestModelMap();
    expect(() => model.stringCodec().decode("a"),
        throwsA(const TypeMatcher<FormatException>()));
  });

  test("==", () {
    final model = TestModelMap();
    final model2 = TestModelMap().clone();
    expect(model, model2);
    model2.boolVar.applyToggle();
    expect(model, isNot(model2));
  });

  test("hashCode", () {
    final model = TestModelMap();
    final model2 = TestModelMap().clone();
    expect(model.hashCode, model2.hashCode);
    model2.boolVar.applyToggle();
    expect(model.hashCode, isNot(model2.hashCode));
  });

  test("debugToString", () {
    final model = TestModelMap();
    expect(model.debugToString(),
        stringContainsInOrder(["$TestModelMap", "toString"]));
  });
}

Map<String, dynamic> testModelMap = <String, dynamic>{
  "bool": true,
  "double": 0.1234,
  "int": -34,
  "string": "ok",
  "stringdict": {
    "dictkey": "value",
  },
  "booldict": {
    "bla": true,
  },
  "emptybooldict": <String, dynamic>{
  },
  "smartdecimal": "123.1231231231231231231231231231231231231293810293812093",
};

class TestModelDict extends SmartJSON<TestModelDict> {
  final SmartModelDict<TestModelMap> modelDict = SmartModelDict(
      {"bla": TestModelMap()}, () => TestModelMap());

  @override TestModelDict creator() => TestModelDict();

  @override JSONPrototypeCodec<TestModelDict> tactic() =>
      JSONMapTactic.ignoreMissing({"model": modelDict});
}

class TestModelMap extends SmartJSON<TestModelMap> {
  final SmartBool boolVar = SmartBool(testModelMap["bool"] as bool);
  final SmartDouble doubleVar = SmartDouble(testModelMap["double"] as double);
  final SmartInt intVar = SmartInt(testModelMap["int"] as int);
  final SmartString stringVar = SmartString(testModelMap["string"] as String);
  final SmartDictT<String> stringDictVar = SmartDictT(
      testModelMap["stringdict"] as Map<String, String>);
  final SmartDictT<bool> boolDictVar = SmartDictT(
      testModelMap["booldict"] as Map<String, bool>);
  final SmartDictT<bool> emptyBoolDictVar = SmartDictT({});
  final SmartLazyDecimal decimal = SmartLazyDecimal(
      "123.1231231231231231231231231231231231231293810293812093");

  @override
  TestModelMap creator() => TestModelMap();

  @override
  JSONPrototypeCodec<TestModelMap> tactic() =>
      JSONMapTactic.ignoreMissing({
        "bool": boolVar,
        "double": doubleVar,
        "int": intVar,
        "string": stringVar,
        "stringdict": stringDictVar,
        "booldict": boolDictVar,
        "emptybooldict": emptyBoolDictVar,
        "smartdecimal": decimal,
      });
}

List<dynamic> testModelList = <dynamic>[true, 0.1234, -34, "ok"];

class _TestModelList extends SmartJSON<_TestModelList> {
  final SmartBool boolVar = SmartBool(testModelList[0] as bool);
  final SmartDouble doubleVar = SmartDouble(testModelList[1] as double);
  final SmartInt intVar = SmartInt(testModelList[2] as int);
  final SmartString stringVar = SmartString(testModelList[3] as String);

  @override
  _TestModelList creator() => _TestModelList();

  @override
  JSONPrototypeCodec<_TestModelList> tactic() =>
      JSONListTactic([
        boolVar,
        doubleVar,
        intVar,
        stringVar,
      ]);
}

class _ListOfLists extends SmartJSON<_ListOfLists> {
  final SmartModelList<_ListData> data = SmartModelList(
      [_ListData(), _ListData()], () => _ListData());

  @override
  _ListOfLists creator() => _ListOfLists();

  @override JSONPrototypeCodec<_ListOfLists> tactic() => JSONSingleTactic(data);
}

List<dynamic> list = <dynamic>[false, 0, 1.0];

class _ListData extends SmartJSON<_ListData> {
  final SmartBool boolVar = SmartBool(list[0] as bool);
  final SmartInt intVar = SmartInt(list[1] as int);
  final SmartDouble doubleVar = SmartDouble(list[2] as double);

  @override _ListData creator() => _ListData();

  @override JSONPrototypeCodec<_ListData> tactic() =>
      JSONListTactic([
        boolVar,
        intVar,
        doubleVar,
      ]);
}

class _MockSmartJSON extends SmartJSON<_MockSmartJSON> {

  final SmartString str = SmartString("b");

  @override
  _MockSmartJSON creator() => _MockSmartJSON();

  @override
  JSONPrototypeCodec<_MockSmartJSON> tactic() {
    return JSONMapTactic.nullOnMissing({
      "a": str,
    });
  }
}

enum _TestEnum { a, b, c }